# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any, Tuple, List
import os

from mautrix.types import UserID
from mautrix.util.config import ConfigUpdateHelper, ForbiddenDefault, ForbiddenKey
from mautrix.bridge.config import BaseBridgeConfig


class Config(BaseBridgeConfig):
    def __getitem__(self, key: str) -> Any:
        try:
            return os.environ[f"MAUTRIX_FACEBOOK_{key.replace('.', '_').upper()}"]
        except KeyError:
            return super().__getitem__(key)

    @property
    def forbidden_defaults(self) -> List[ForbiddenDefault]:
        return [
            *super().forbidden_defaults,
            ForbiddenDefault("appservice.database", "postgres://username:password@hostname/db"),
            ForbiddenDefault("appservice.public.external", "https://example.com/public",
                             condition="appservice.public.enabled"),
            ForbiddenDefault("bridge.permissions", ForbiddenKey("example.com"))
        ]

    def do_update(self, helper: ConfigUpdateHelper) -> None:
        super().do_update(helper)

        copy, copy_dict, base = helper

        copy("homeserver.asmux")

        copy("appservice.community_id")

        if self["appservice.bot_avatar"] == "mxc://maunium.net/ddtNPZSKMNqaUzqrHuWvUADv":
            base["appservice.bot_avatar"] = "mxc://maunium.net/ygtkteZsXnGJLJHRchUwYWak"

        copy("appservice.public.enabled")
        copy("appservice.public.prefix")
        copy("appservice.public.external")
        if self["appservice.public.shared_secret"] == "generate":
            base["appservice.public.shared_secret"] = self._new_token()
        else:
            copy("appservice.public.shared_secret")

        copy("metrics.enabled")
        copy("metrics.listen_port")

        copy("bridge.username_template")
        copy("bridge.displayname_template")
        copy("bridge.displayname_preference")
        copy("bridge.community_template")
        copy("bridge.command_prefix")

        copy("bridge.initial_chat_sync")
        copy("bridge.invite_own_puppet_to_pm")
        copy("bridge.sync_with_custom_puppets")
        copy("bridge.sync_direct_chat_list")
        copy("bridge.double_puppet_server_map")
        copy("bridge.double_puppet_allow_discovery")
        if "bridge.login_shared_secret" in self:
            base["bridge.login_shared_secret_map"] = {
                base["homeserver.domain"]: self["bridge.login_shared_secret"]
            }
        else:
            copy("bridge.login_shared_secret_map")
        copy("bridge.presence")
        copy("bridge.update_avatar_initial_sync")
        copy("bridge.encryption.allow")
        copy("bridge.encryption.default")
        copy("bridge.encryption.key_sharing.allow")
        copy("bridge.encryption.key_sharing.require_cross_signing")
        copy("bridge.encryption.key_sharing.require_verification")
        copy("bridge.delivery_receipts")
        copy("bridge.allow_invites")
        copy("bridge.backfill.invite_own_puppet")
        copy("bridge.backfill.initial_limit")
        copy("bridge.backfill.missed_limit")
        copy("bridge.backfill.disable_notifications")
        if "bridge.periodic_reconnect_interval" in self:
            base["bridge.periodic_reconnect.interval"] = self["bridge.periodic_reconnect_interval"]
            base["bridge.periodic_reconnect.mode"] = self["bridge.periodic_reconnect_mode"]
        else:
            copy("bridge.periodic_reconnect.interval")
            copy("bridge.periodic_reconnect.mode")
            copy("bridge.periodic_reconnect.always")
        copy("bridge.resync_max_disconnected_time")
        copy("bridge.temporary_disconnect_notices")
        copy("bridge.refresh_on_reconnection_fail")
        copy("bridge.resend_bridge_info")

        copy_dict("bridge.permissions")

    def _get_permissions(self, key: str) -> Tuple[bool, bool, str]:
        level = self["bridge.permissions"].get(key, "")
        admin = level == "admin"
        user = level == "user" or admin
        return user, admin, level

    def get_permissions(self, mxid: UserID) -> Tuple[bool, bool, str]:
        permissions = self["bridge.permissions"] or {}
        if mxid in permissions:
            return self._get_permissions(mxid)

        homeserver = mxid[mxid.index(":") + 1:]
        if homeserver in permissions:
            return self._get_permissions(homeserver)

        return self._get_permissions("*")



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Dict
import base64
import struct
import time
import io

from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_v1_5, AES
from Crypto.Random import get_random_bytes

from .base import BaseAndroidAPI
from .errors import TwoFactorRequired
from ..types import LoginResponse, PasswordKeyResponse, MobileConfig


class LoginAPI(BaseAndroidAPI):
    async def pwd_key_fetch(self) -> PasswordKeyResponse:
        req_data = self.format({
            "version": "2",
            "flow": "CONTROLLER_INITIALIZATION",
            **self._params,
            "method": "GET",
            "fb_api_req_friendly_name": "pwdKeyFetch",
            "fb_api_caller_class": "com.facebook.auth.login.AuthOperations",
            "access_token": self.state.application.access_token,
        }, sign=False)
        resp = await self.http.post(self.graph_url.with_path("//pwd_key_fetch"),
                                    headers=self._headers, data=req_data)
        json_data = await self._handle_response(resp)
        parsed = PasswordKeyResponse.deserialize(json_data)
        self.state.session.password_encryption_pubkey = parsed.public_key
        self.state.session.password_encryption_key_id = parsed.key_id
        return parsed

    async def mobile_config_sessionless(self) -> MobileConfig:
        req_data = self.format({
            "query_hash": "4d8d81394cc798f2a72f5761248743fa7f4d3c6b0048f7e9696006230f118896",
            "one_query_hash": "b1f0cf90cfb2d8424ecb5f82c9d9365a03bf7f3af3315df1c8d9f0b5e4818bdd",
            "bool_opt_policy": "1",
            "device_id": self.state.device.uuid,
            "api_version": "7",
            "name_to_id": "true",
            "fetch_type": "SYNC_FULL",
            "access_token": self.state.application.access_token,
            **self._params,
        }, sign=False)
        resp = await self.http.post(self.b_graph_url / "mobileconfigsessionless",
                                    headers=self._headers, data=req_data)
        json_data = await self._handle_response(resp)
        parsed = MobileConfig.deserialize(json_data)
        self.state.session.password_encryption_key_id = parsed.find(15712, 1).i64
        self.state.session.password_encryption_pubkey = parsed.find(15712, 2).str
        return parsed

    async def login(self, email: str, password: Optional[str] = None,
                    encrypted_password: Optional[str] = None) -> LoginResponse:
        if password:
            if encrypted_password:
                raise ValueError("Only one of password or encrypted_password must be provided")
            encrypted_password = self._encrypt_password(password)
        elif not encrypted_password:
            raise ValueError("One of password or encrypted_password is required")
        return await self._login(email=email, password=encrypted_password,
                                 credentials_type="password")

    async def login_2fa(self, email: str, code: str) -> LoginResponse:
        if not self.state.session.login_first_factor:
            raise ValueError("No two-factor login in progress")
        return await self._login(email=email, password=code, twofactor_code=code,
                                 encrypted_msisdn="", currently_logged_in_userid="0",
                                 userid=str(self.state.session.uid),
                                 machine_id=self.state.session.machine_id,
                                 first_factor=self.state.session.login_first_factor,
                                 credentials_type="two_factor")

    async def login_approved(self) -> LoginResponse:
        if not self.state.session.transient_auth_token:
            raise ValueError("No two-factor login in progress")
        return await self._login(password=self.state.session.transient_auth_token,
                                 email=str(self.state.session.uid), encrypted_msisdn="",
                                 credentials_type="transient_token")

    async def check_approved_machine(self) -> bool:
        req: Dict[str, str] = {
            "u": str(self.state.session.uid),
            "m": self.state.session.machine_id,
            **self._params,
            "method": "GET",
            "fb_api_req_friendly_name": "checkApprovedMachine",
            "fb_api_caller_class": "com.facebook.account.twofac.protocol.TwoFacServiceHandler",
            "access_token": self.state.application.access_token,
        }
        headers = {
            **self._headers,
            "content-type": "application/x-www-form-urlencoded",
            "x-fb-friendly-name": req["fb_api_req_friendly_name"],
        }
        resp = await self.http.post(url=self.graph_url / "check_approved_machine",
                                    headers=headers, data=req)
        json_data = await self._handle_response(resp)
        return json_data["data"][0]["approved"]

    async def _login(self, **kwargs) -> LoginResponse:
        req: Dict[str, str] = {
            **self._params,
            "adid": self.state.device.adid,
            "api_key": self.state.application.client_id,
            "community_id": "",
            "cpl": "true",
            "currently_logged_in_userid": "0",
            "device_id": self.state.device.uuid,
            "fb_api_caller_class": "com.facebook.auth.login.AuthOperations$PasswordAuthOperation",
            "fb_api_req_friendly_name": "authenticate",
            "format": "json",
            "generate_analytics_claim": "1",
            "generate_machine_id": "1",
            "generate_session_cookies": "1",
            "jazoest": self._jazoest,
            "meta_inf_fbmeta": "NO_FILE",
            "source": "login",
            "try_num": "1",  # TODO maybe cache this somewhere?
            **kwargs,
        }
        req_data = self.format(req, sign=True, access_token=self.state.application.access_token)
        headers = {
            **self._headers,
            "content-type": "application/x-www-form-urlencoded",
            "x-fb-friendly-name": req["fb_api_req_friendly_name"],
        }
        resp = await self.http.post(url=self.b_graph_url / "auth" / "login",
                                    headers=headers, data=req_data)
        self.log.trace(f"Login response: {resp.status} {await resp.text()}")
        try:
            json_data = await self._handle_response(resp)
        except TwoFactorRequired as e:
            self.state.session.machine_id = e.machine_id
            self.state.session.uid = e.uid
            self.state.session.login_first_factor = e.login_first_factor
            self.state.session.transient_auth_token = e.auth_token
            raise
        parsed = LoginResponse.deserialize(json_data)
        self.state.session.access_token = parsed.access_token
        self.state.session.uid = parsed.uid
        self.state.session.machine_id = parsed.machine_id
        self.state.session.login_first_factor = None
        # TODO maybe store the cookies and other data too?
        return parsed

    def _encrypt_password(self, password: str) -> str:
        # Key and IV for AES encryption
        rand_key = get_random_bytes(32)
        iv = get_random_bytes(12)

        # Encrypt AES key with Facebook's RSA public key
        pubkey_bytes = self.state.session.password_encryption_pubkey
        pubkey = RSA.import_key(pubkey_bytes)
        cipher_rsa = PKCS1_v1_5.new(pubkey)
        encrypted_rand_key = cipher_rsa.encrypt(rand_key)

        cipher_aes = AES.new(rand_key, AES.MODE_GCM, nonce=iv)
        # Add the current time to the additional authenticated data (AAD) section
        current_time = int(time.time())
        cipher_aes.update(str(current_time).encode("utf-8"))
        # Encrypt the password and get the AES MAC auth tag
        encrypted_passwd, auth_tag = cipher_aes.encrypt_and_digest(password.encode("utf-8"))

        buf = io.BytesIO()
        # 1 is presumably the version
        buf.write(bytes([1, int(self.state.session.password_encryption_key_id)]))
        buf.write(iv)
        # Length of the encrypted AES key as a little-endian 16-bit int
        buf.write(struct.pack("<h", len(encrypted_rand_key)))
        buf.write(encrypted_rand_key)
        buf.write(auth_tag)
        buf.write(encrypted_passwd)
        encoded = base64.b64encode(buf.getvalue()).decode("utf-8")
        return f"#PWD_MSGR:1:{current_time}:{encoded}"

    @property
    def _jazoest(self) -> str:
        return f"2{sum(ord(i) for i in self.state.device.uuid)}"



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Union, List, TYPE_CHECKING, ClassVar
from enum import Enum

from asyncpg import Record
from attr import dataclass

from mautrix.types import RoomID, ContentURI
from mautrix.util.async_db import Database
from maufbapi.types.mqtt import ThreadKey as MQTTThreadKey
from maufbapi.types.graphql import ThreadKey as GraphQLThreadKey

fake_db = Database("") if TYPE_CHECKING else None


class ThreadType(Enum):
    USER = "USER"
    GROUP = "GROUP"
    PAGE = "PAGE"
    UNKNOWN = "UNKNOWN"

    @classmethod
    def from_thread_key(cls, thread_key: Union[MQTTThreadKey, GraphQLThreadKey]) -> 'ThreadType':
        if thread_key.thread_fbid:
            return cls.GROUP
        elif thread_key.other_user_id:
            return cls.USER
        else:
            return cls.UNKNOWN


@dataclass
class Portal:
    db: ClassVar[Database] = fake_db

    fbid: int
    fb_receiver: int
    fb_type: ThreadType
    mxid: Optional[RoomID]
    name: Optional[str]
    photo_id: Optional[str]
    avatar_url: Optional[ContentURI]
    encrypted: bool

    @classmethod
    def _from_row(cls, row: Optional[Record]) -> Optional['Portal']:
        if row is None:
            return None
        data = {**row}
        fb_type = ThreadType(data.pop("fb_type"))
        return cls(**data, fb_type=fb_type)

    @classmethod
    async def get_by_fbid(cls, fbid: int, fb_receiver: int) -> Optional['Portal']:
        q = ("SELECT fbid, fb_receiver, fb_type, mxid, name, photo_id, avatar_url, encrypted "
             "FROM portal WHERE fbid=$1 AND fb_receiver=$2")
        row = await cls.db.fetchrow(q, fbid, fb_receiver)
        return cls._from_row(row)

    @classmethod
    async def get_by_mxid(cls, mxid: RoomID) -> Optional['Portal']:
        q = ("SELECT fbid, fb_receiver, fb_type, mxid, name, photo_id, avatar_url, encrypted "
             "FROM portal WHERE mxid=$1")
        row = await cls.db.fetchrow(q, mxid)
        return cls._from_row(row)

    @classmethod
    async def get_all_by_receiver(cls, fb_receiver: int) -> List['Portal']:
        q = ("SELECT fbid, fb_receiver, fb_type, mxid, name, photo_id, avatar_url, encrypted "
             "FROM portal WHERE fb_receiver=$1 AND fb_type='USER'")
        rows = await cls.db.fetch(q, fb_receiver)
        return [cls._from_row(row) for row in rows]

    @classmethod
    async def all(cls) -> List['Portal']:
        q = ("SELECT fbid, fb_receiver, fb_type, mxid, name, photo_id, avatar_url, encrypted "
             "FROM portal")
        rows = await cls.db.fetch(q)
        return [cls._from_row(row) for row in rows]

    async def insert(self) -> None:
        q = ("INSERT INTO portal (fbid, fb_receiver, fb_type, mxid, name, photo_id, avatar_url, "
             "                    encrypted) "
             "VALUES ($1, $2, $3, $4, $5, $6, $7, $8)")
        await self.db.execute(q, self.fbid, self.fb_receiver, self.fb_type.name, self.mxid,
                              self.name, self.photo_id, self.avatar_url, self.encrypted)

    async def delete(self) -> None:
        q = "DELETE FROM portal WHERE fbid=$1 AND fb_receiver=$2"
        await self.db.execute(q, self.fbid, self.fb_receiver)

    async def save(self) -> None:
        await self.db.execute("UPDATE portal SET mxid=$3, name=$4, photo_id=$5, avatar_url=$6,"
                              "                  encrypted=$7 "
                              "WHERE fbid=$1 AND fb_receiver=$2",
                              self.fbid, self.fb_receiver, self.mxid, self.name, self.photo_id,
                              self.avatar_url, self.encrypted)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import paho.mqtt.client
import struct
# from hyperframe.frame import SettingsFrame, HeadersFrame, DataFrame
# import hpack

http2_header = b"PRI * HTTP/2.0\r\n\r\nSM\r\n\r\n"


class MQTToTClient(paho.mqtt.client.Client):
    def set_client_id(self, client_id: bytes) -> None:
        self._client_id = client_id

    # This is equivalent to the original _send_connect, except:
    # * the protocol ID is MQTToT.
    # * the client ID is sent without a length.
    # * all extra stuff like wills, usernames, passwords and MQTTv5 is removed.
    def _send_connect(self, keepalive):
        proto_ver = self._protocol
        protocol = b"MQTToT"

        remaining_length = (2 + len(protocol) + 1 +
                            1 + 2 + len(self._client_id))

        connect_flags = 0x02

        command = paho.mqtt.client.CONNECT
        packet = bytearray()
        packet.append(command)

        self._pack_remaining_length(packet, remaining_length)
        packet.extend(struct.pack(f"!H{len(protocol)}sBBH",
                                  len(protocol), protocol, proto_ver, connect_flags, keepalive))
        packet.extend(self._client_id)

        # settings = SettingsFrame(settings={
        #     SettingsFrame.HEADER_TABLE_SIZE: 4096,
        #     SettingsFrame.ENABLE_PUSH: 0,
        #     SettingsFrame.MAX_FRAME_SIZE: 16834,
        #     SettingsFrame.MAX_CONCURRENT_STREAMS: 100,
        # })
        # headers = HeadersFrame(stream_id=1, data=hpack.Encoder().encode({
        #     ":method": "POST",
        #     ":scheme": "http",
        #     ":path": "",
        #     ":authority": "edge-mqtt-merge.facebook.com",
        # }))
        # data = DataFrame(stream_id=1, data=packet)
        # packet = http2_header + settings.serialize() + headers.serialize() + data.serialize()

        self._keepalive = keepalive
        self._easy_log(
            paho.mqtt.client.MQTT_LOG_DEBUG,
            "Sending CONNECT",
        )
        return self._packet_queue(command, packet, 0, 0)



"""Stop using size-limited string fields

Revision ID: 4face019555a
Revises: 3d4f3a252006
Create Date: 2020-11-10 01:43:55.557931

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '4face019555a'
down_revision = '3d4f3a252006'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().engine.name != "postgresql":
        return

    op.alter_column("portal", "fbid", type_=sa.Text())
    op.alter_column("portal", "fb_receiver", type_=sa.Text())
    op.alter_column("portal", "mxid", type_=sa.Text())
    op.alter_column("portal", "avatar_url", type_=sa.Text())
    op.alter_column("portal", "name", type_=sa.Text())
    op.alter_column("portal", "photo_id", type_=sa.Text())

    op.alter_column("message", "mxid", type_=sa.Text())
    op.alter_column("message", "mx_room", type_=sa.Text())
    op.alter_column("message", "fbid", type_=sa.Text())
    op.alter_column("message", "fb_chat", type_=sa.Text())
    op.alter_column("message", "fb_receiver", type_=sa.Text())

    op.alter_column("puppet", "fbid", type_=sa.Text())
    op.alter_column("puppet", "name", type_=sa.Text())
    op.alter_column("puppet", "photo_id", type_=sa.Text())
    op.alter_column("puppet", "custom_mxid", type_=sa.Text())
    op.alter_column("puppet", "next_batch", type_=sa.Text())

    op.alter_column("reaction", "mxid", type_=sa.Text())
    op.alter_column("reaction", "mx_room", type_=sa.Text())
    op.alter_column("reaction", "fb_msgid", type_=sa.Text())
    op.alter_column("reaction", "fb_receiver", type_=sa.Text())
    op.alter_column("reaction", "fb_sender", type_=sa.Text())
    op.alter_column("reaction", "reaction", type_=sa.Text())

    op.alter_column("user", "mxid", type_=sa.Text())
    op.alter_column("user", "fbid", type_=sa.Text())
    op.alter_column("user", "notice_room", type_=sa.Text())
    op.alter_column("user", "user_agent", type_=sa.Text())
    op.alter_column("user", "fb_domain", type_=sa.Text())

    op.alter_column("user_portal", "user", type_=sa.Text())
    op.alter_column("user_portal", "portal", type_=sa.Text())
    op.alter_column("user_portal", "portal_receiver", type_=sa.Text())

    op.alter_column("contact", "user", type_=sa.Text())
    op.alter_column("contact", "contact", type_=sa.Text())

    op.alter_column("mx_user_profile", "room_id", type_=sa.Text())
    op.alter_column("mx_user_profile", "user_id", type_=sa.Text())
    op.alter_column("mx_user_profile", "displayname", type_=sa.Text())
    op.alter_column("mx_user_profile", "avatar_url", type_=sa.Text())
    op.alter_column("mx_room_state", "room_id", type_=sa.Text())


def downgrade():
    if op.get_bind().engine.name != "postgresql":
        return

    op.alter_column("portal", "fbid", type_=sa.String(127))
    op.alter_column("portal", "fb_receiver", type_=sa.String(127))
    op.alter_column("portal", "mxid", type_=sa.String(255))
    op.alter_column("portal", "avatar_url", type_=sa.String(255))
    op.alter_column("portal", "name", type_=sa.String(255))
    op.alter_column("portal", "photo_id", type_=sa.String(255))

    op.alter_column("message", "mxid", type_=sa.String(255))
    op.alter_column("message", "mx_room", type_=sa.String(255))
    op.alter_column("message", "fbid", type_=sa.String(127))
    op.alter_column("message", "fb_chat", type_=sa.String(127))
    op.alter_column("message", "fb_receiver", type_=sa.String(127))

    op.alter_column("puppet", "fbid", type_=sa.String(127))
    op.alter_column("puppet", "name", type_=sa.String(255))
    op.alter_column("puppet", "photo_id", type_=sa.String(255))
    op.alter_column("puppet", "custom_mxid", type_=sa.String(255))
    op.alter_column("puppet", "next_batch", type_=sa.String(255))

    op.alter_column("reaction", "mxid", type_=sa.String(255))
    op.alter_column("reaction", "mx_room", type_=sa.String(255))
    op.alter_column("reaction", "fb_msgid", type_=sa.String(127))
    op.alter_column("reaction", "fb_receiver", type_=sa.String(127))
    op.alter_column("reaction", "fb_sender", type_=sa.String(127))
    op.alter_column("reaction", "reaction", type_=sa.String(1))

    op.alter_column("user", "mxid", type_=sa.String(255))
    op.alter_column("user", "fbid", type_=sa.String(255))
    op.alter_column("user", "notice_room", type_=sa.String(255))
    op.alter_column("user", "user_agent", type_=sa.String(255))
    op.alter_column("user", "fb_domain", type_=sa.String(255))

    op.alter_column("user_portal", "user", type_=sa.String(255))
    op.alter_column("user_portal", "portal", type_=sa.String(255))
    op.alter_column("user_portal", "portal_receiver", type_=sa.String(255))

    op.alter_column("contact", "user", type_=sa.String(255))
    op.alter_column("contact", "contact", type_=sa.String(255))

    op.alter_column("mx_user_profile", "room_id", type_=sa.String(255))
    op.alter_column("mx_user_profile", "user_id", type_=sa.String(255))
    op.alter_column("mx_user_profile", "displayname", type_=sa.String())
    op.alter_column("mx_user_profile", "avatar_url", type_=sa.String(255))
    op.alter_column("mx_room_state", "room_id", type_=sa.String(255))



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from mautrix.bridge.commands import HelpSection, command_handler

from .typehint import CommandEvent

SECTION_CONNECTION = HelpSection("Connection management", 15, "")


@command_handler(needs_auth=False, management_only=True, help_section=SECTION_CONNECTION,
                 help_text="Mark this room as your bridge notice room")
async def set_notice_room(evt: CommandEvent) -> None:
    evt.sender.notice_room = evt.room_id
    await evt.sender.save()
    await evt.reply("This room has been marked as your bridge notice room")


@command_handler(needs_auth=True, management_only=True, help_section=SECTION_CONNECTION,
                 help_text="Disconnect from Facebook Messenger")
async def disconnect(evt: CommandEvent) -> None:
    if not evt.sender.mqtt:
        await evt.reply("You don't have a Messenger MQTT connection")
        return
    evt.sender.mqtt.disconnect()


@command_handler(needs_auth=True, management_only=True, help_section=SECTION_CONNECTION,
                 help_text="Connect to Facebook Messenger", aliases=["reconnect"])
async def connect(evt: CommandEvent) -> None:
    if evt.sender.listen_task and not evt.sender.listen_task.done():
        await evt.reply("You already have a Messenger MQTT connection")
        return
    evt.sender.start_listen()


@command_handler(needs_auth=True, management_only=True, help_section=SECTION_CONNECTION,
                 help_text="Check if you're logged into Facebook Messenger")
async def ping(evt: CommandEvent) -> None:
    if not await evt.sender.is_logged_in():
        await evt.reply("You're not logged into Facebook Messenger")
        return
    # try:
    own_info = await evt.sender.client.get_self()
    # TODO catch errors
    # except fbchat.PleaseRefresh as e:
    #     await evt.reply(f"{e}\n\nUse `$cmdprefix+sp refresh` refresh the session.")
    #     return
    await evt.reply(f"You're logged in as {own_info.name} (user ID {own_info.id})")

    if not evt.sender.listen_task or evt.sender.listen_task.done():
        await evt.reply("You don't have a Messenger MQTT connection. Use `connect` to connect.")
    elif not evt.sender.is_connected:
        await evt.reply("The Messenger MQTT listener is **disconnected**.")
    else:
        await evt.reply("The Messenger MQTT listener is connected.")


@command_handler(needs_auth=True, management_only=True, help_section=SECTION_CONNECTION,
                 help_text="\"Refresh\" the Facebook Messenger page")
async def refresh(evt: CommandEvent) -> None:
    await evt.sender.refresh(force_notice=True)



"""Add end-to-bridge encryption fields

Revision ID: c56c9a30b228
Revises: 44a02762423d
Create Date: 2020-04-05 01:06:06.277403

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'c56c9a30b228'
down_revision = '44a02762423d'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('nio_account',
    sa.Column('user_id', sa.String(length=255), nullable=False),
    sa.Column('device_id', sa.String(length=255), nullable=False),
    sa.Column('shared', sa.Boolean(), nullable=False),
    sa.Column('sync_token', sa.Text(), nullable=False),
    sa.Column('account', sa.LargeBinary(), nullable=False),
    sa.PrimaryKeyConstraint('user_id', 'device_id')
    )
    op.create_table('nio_device_key',
    sa.Column('user_id', sa.String(length=255), nullable=False),
    sa.Column('device_id', sa.String(length=255), nullable=False),
    sa.Column('display_name', sa.String(length=255), nullable=False),
    sa.Column('deleted', sa.Boolean(), nullable=False),
    sa.Column('keys', sa.PickleType(), nullable=False),
    sa.PrimaryKeyConstraint('user_id', 'device_id')
    )
    op.create_table('nio_megolm_inbound_session',
    sa.Column('session_id', sa.String(length=255), nullable=False),
    sa.Column('sender_key', sa.String(length=255), nullable=False),
    sa.Column('fp_key', sa.String(length=255), nullable=False),
    sa.Column('room_id', sa.String(length=255), nullable=False),
    sa.Column('session', sa.LargeBinary(), nullable=False),
    sa.Column('forwarded_chains', sa.PickleType(), nullable=False),
    sa.PrimaryKeyConstraint('session_id')
    )
    op.create_table('nio_olm_session',
    sa.Column('session_id', sa.String(length=255), nullable=False),
    sa.Column('sender_key', sa.String(length=255), nullable=False),
    sa.Column('session', sa.LargeBinary(), nullable=False),
    sa.Column('created_at', sa.DateTime(), nullable=False),
    sa.Column('last_used', sa.DateTime(), nullable=False),
    sa.PrimaryKeyConstraint('session_id')
    )
    op.create_table('nio_outgoing_key_request',
    sa.Column('request_id', sa.String(length=255), nullable=False),
    sa.Column('session_id', sa.String(length=255), nullable=False),
    sa.Column('room_id', sa.String(length=255), nullable=False),
    sa.Column('algorithm', sa.String(length=255), nullable=False),
    sa.PrimaryKeyConstraint('request_id')
    )
    with op.batch_alter_table('portal') as batch_op:
        batch_op.add_column(sa.Column('encrypted', sa.Boolean(), server_default=sa.false(), nullable=False))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('portal') as batch_op:
        batch_op.drop_column('encrypted')
    op.drop_table('nio_outgoing_key_request')
    op.drop_table('nio_olm_session')
    op.drop_table('nio_megolm_inbound_session')
    op.drop_table('nio_device_key')
    op.drop_table('nio_account')
    # ### end Alembic commands ###



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import (Optional, Dict, AsyncGenerator, AsyncIterable, Awaitable, Union, TYPE_CHECKING,
                    cast)
from datetime import datetime, timedelta
import asyncio

from yarl import URL
import magic

from mautrix.types import UserID, RoomID, SyncToken, ContentURI
from mautrix.appservice import IntentAPI
from mautrix.bridge import BasePuppet, async_getter_lock
from mautrix.util.simple_template import SimpleTemplate
from maufbapi.types.graphql import Participant, Picture

from .config import Config
from .db import Puppet as DBPuppet
from . import user as u, portal as p, matrix as m

if TYPE_CHECKING:
    from .__main__ import MessengerBridge


class Puppet(DBPuppet, BasePuppet):
    mx: m.MatrixHandler
    config: Config
    hs_domain: str
    mxid_template: SimpleTemplate[int]

    by_fbid: Dict[int, 'Puppet'] = {}
    by_custom_mxid: Dict[UserID, 'Puppet'] = {}

    _last_info_sync: Optional[datetime]

    def __init__(self, fbid: int, name: Optional[str] = None, photo_id: Optional[str] = None,
                 photo_mxc: Optional[ContentURI] = None, name_set: bool = False,
                 avatar_set: bool = False, is_registered: bool = False,
                 custom_mxid: Optional[UserID] = None, access_token: Optional[str] = None,
                 next_batch: Optional[SyncToken] = None, base_url: Optional[URL] = None) -> None:
        super().__init__(fbid, name, photo_id, photo_mxc, name_set, avatar_set, is_registered,
                         custom_mxid, access_token, next_batch, base_url)
        self._last_info_sync = None

        self.default_mxid = self.get_mxid_from_id(fbid)
        self.default_mxid_intent = self.az.intent.user(self.default_mxid)
        self.intent = self._fresh_intent()

        self.log = self.log.getChild(str(self.fbid))

    @property
    def should_sync(self) -> bool:
        now = datetime.now()
        if not self._last_info_sync or now - self._last_info_sync > timedelta(hours=48):
            self._last_info_sync = now
            return True
        return False

    async def default_puppet_should_leave_room(self, room_id: RoomID) -> bool:
        portal = await p.Portal.get_by_mxid(room_id)
        return portal and portal.fbid != self.fbid

    async def _leave_rooms_with_default_user(self) -> None:
        await super()._leave_rooms_with_default_user()
        # Make the user join all private chat portals.
        await asyncio.gather(*[self.intent.ensure_joined(portal.mxid)
                               async for portal in p.Portal.get_all_by_receiver(self.fbid)
                               if portal.mxid], loop=self.loop)

    def intent_for(self, portal: 'p.Portal') -> IntentAPI:
        if portal.fbid == self.fbid or (portal.backfill_lock.locked
                                        and self.config["bridge.backfill.invite_own_puppet"]):
            return self.default_mxid_intent
        return self.intent

    @classmethod
    def init_cls(cls, bridge: 'MessengerBridge') -> AsyncIterable[Awaitable[None]]:
        cls.config = bridge.config
        cls.loop = bridge.loop
        cls.mx = bridge.matrix
        cls.az = bridge.az
        cls.hs_domain = cls.config["homeserver.domain"]
        cls.mxid_template = SimpleTemplate(cls.config["bridge.username_template"], "userid",
                                           prefix="@", suffix=f":{Puppet.hs_domain}", type=int)
        cls.sync_with_custom_puppets = cls.config["bridge.sync_with_custom_puppets"]
        cls.homeserver_url_map = {server: URL(url) for server, url
                                  in cls.config["bridge.double_puppet_server_map"].items()}
        cls.allow_discover_url = cls.config["bridge.double_puppet_allow_discovery"]
        cls.login_shared_secret_map = {server: secret.encode("utf-8") for server, secret
                                       in cls.config["bridge.login_shared_secret_map"].items()}
        cls.login_device_name = "Facebook Messenger Bridge"

        return (puppet.try_start() async for puppet in Puppet.get_all_with_custom_mxid())

    # region User info updating

    async def update_info(self, source: Optional['u.User'] = None, info: Participant = None,
                          update_avatar: bool = True) -> 'Puppet':
        if not info:
            # if not self.should_sync:
            #     return self
            # FIXME
            # info = await source.client.fetch_thread_info([self.fbid]).__anext__()
            print("no info to update puppet :(")
            return self
        self._last_info_sync = datetime.now()
        try:
            changed = await self._update_name(info)
            if update_avatar:
                changed = await self._update_photo(source, info.profile_pic_large) or changed
            if changed:
                await self.save()
        except Exception:
            self.log.exception(f"Failed to update info from source {source.fbid}")
        return self

    @classmethod
    def _get_displayname(cls, info: Participant) -> str:
        sn = info.structured_name
        info = {
            "displayname": None,
            "id": info.id,
            "name": info.name,
            "phonetic_name": sn.phonetic_name if sn else None,
            "own_nickname": info.nickname_for_viewer,
            **(sn.to_dict() if sn else {}),
        }
        for preference in cls.config["bridge.displayname_preference"]:
            if info.get(preference):
                info["displayname"] = info.get(preference)
                break
        return cls.config["bridge.displayname_template"].format(**info)

    async def _update_name(self, info: Participant) -> bool:
        name = self._get_displayname(info)
        if name != self.name or not self.name_set:
            self.name = name
            try:
                await self.default_mxid_intent.set_displayname(self.name)
                self.name_set = True
            except Exception:
                self.log.exception("Failed to set displayname")
                self.name_set = False
            return True
        return False

    @staticmethod
    async def reupload_avatar(source: Optional['u.User'], intent: IntentAPI, url: str,
                              fbid: int, use_graph: bool = True) -> ContentURI:
        data = None
        if use_graph and source and source.state and source.state.session.access_token:
            graph_url = ((source.client.graph_url / str(fbid) / "picture")
                         .with_query({"width": "1000", "height": "1000"}))
            async with source.client.get(graph_url) as resp:
                if resp.status < 400:
                    data = await resp.read()
        if data is None:
            async with source.client.get(url) as resp:
                data = await resp.read()
        mime = magic.from_buffer(data, mime=True)
        return await intent.upload_media(data, mime_type=mime)

    async def _update_photo(self, source: 'u.User', photo: Picture) -> bool:
        photo_id = p.Portal.get_photo_id(photo)
        if photo_id != self.photo_id or not self.avatar_set:
            self.photo_id = photo_id
            if photo:
                self.photo_mxc = await self.reupload_avatar(source, self.default_mxid_intent,
                                                            photo.uri, self.fbid,
                                                            use_graph=(photo.height or 0) < 500)
            else:
                self.photo_mxc = ContentURI("")
            try:
                await self.default_mxid_intent.set_avatar_url(self.photo_mxc)
                self.avatar_set = True
            except Exception:
                self.log.exception("Failed to set avatar")
                self.avatar_set = False
            return True
        return False

    # endregion
    # region Database getters

    def _add_to_cache(self) -> None:
        self.by_fbid[self.fbid] = self
        if self.custom_mxid:
            self.by_custom_mxid[self.custom_mxid] = self

    @classmethod
    @async_getter_lock
    async def get_by_fbid(cls, fbid: Union[str, int], *, create: bool = True) -> Optional['Puppet']:
        if isinstance(fbid, str):
            fbid = int(fbid)
        try:
            return cls.by_fbid[fbid]
        except KeyError:
            pass

        puppet = cast(cls, await super().get_by_fbid(fbid))
        if puppet:
            puppet._add_to_cache()
            return puppet

        if create:
            puppet = cls(fbid)
            await puppet.insert()
            puppet._add_to_cache()
            return puppet

        return None

    @classmethod
    async def get_by_mxid(cls, mxid: UserID, create: bool = True) -> Optional['Puppet']:
        fbid = cls.get_id_from_mxid(mxid)
        if fbid:
            return await cls.get_by_fbid(fbid, create=create)
        return None

    @classmethod
    @async_getter_lock
    async def get_by_custom_mxid(cls, mxid: UserID) -> Optional['Puppet']:
        try:
            return cls.by_custom_mxid[mxid]
        except KeyError:
            pass

        puppet = cast(cls, await super().get_by_custom_mxid(mxid))
        if puppet:
            puppet._add_to_cache()
            return puppet

        return None

    @classmethod
    def get_id_from_mxid(cls, mxid: UserID) -> Optional[int]:
        return cls.mxid_template.parse(mxid)

    @classmethod
    def get_mxid_from_id(cls, fbid: int) -> UserID:
        return UserID(cls.mxid_template.format_full(fbid))

    @classmethod
    async def get_all_with_custom_mxid(cls) -> AsyncGenerator['Puppet', None]:
        puppets = await super().get_all_with_custom_mxid()
        puppet: cls
        for puppet in puppets:
            try:
                yield cls.by_fbid[puppet.fbid]
            except KeyError:
                puppet._add_to_cache()
                yield puppet

    # endregion



"""Store user agent in db

Revision ID: 44a02762423d
Revises: 3355b59b191c
Create Date: 2019-08-20 18:43:49.065177

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '44a02762423d'
down_revision = '3355b59b191c'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("user") as batch_op:
        batch_op.add_column(sa.Column("user_agent", sa.String(length=255), nullable=True))


def downgrade():
    with op.batch_alter_table("user") as batch_op:
        batch_op.drop_column("user_agent")



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from mautrix.util.logging.color import ColorFormatter as BaseColorFormatter, PREFIX, RESET

FBCHAT_COLOR = PREFIX + "35;1m"  # magenta


class ColorFormatter(BaseColorFormatter):
    def _color_name(self, module: str) -> str:
        if module.startswith("maufbapi"):
            return FBCHAT_COLOR + module + RESET
        return super()._color_name(module)



"""Store encryption state event in db

Revision ID: 2162f7ae5365
Revises: 905a91536747
Create Date: 2020-07-11 21:12:01.506978

"""
from alembic import op
import sqlalchemy as sa

from mautrix.client.state_store.sqlalchemy import SerializableType
from mautrix.types import RoomEncryptionStateEventContent

# revision identifiers, used by Alembic.
revision = '2162f7ae5365'
down_revision = '905a91536747'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mx_room_state', schema=None) as batch_op:
        batch_op.add_column(sa.Column('encryption',
                                      SerializableType(RoomEncryptionStateEventContent),
                                      nullable=True))
        batch_op.add_column(sa.Column('has_full_member_list', sa.Boolean(), nullable=True))
        batch_op.add_column(sa.Column('is_encrypted', sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mx_room_state', schema=None) as batch_op:
        batch_op.drop_column('is_encrypted')
        batch_op.drop_column('has_full_member_list')
        batch_op.drop_column('encryption')
    # ### end Alembic commands ###



"""Re-add user agent to db

Revision ID: 244273a86d81
Revises: 54aecbd6187c
Create Date: 2020-06-30 15:03:18.260742

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '244273a86d81'
down_revision = '54aecbd6187c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.add_column(sa.Column('user_agent', sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.drop_column('user_agent')

    # ### end Alembic commands ###



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import cast, NamedTuple, List

from mautrix.types import TextMessageEventContent, Format, RoomID, RelationType
from mautrix.util.formatter import (MatrixParser as BaseMatrixParser, MarkdownString, EntityString,
                                    SimpleEntity, EntityType)
from maufbapi.types.mqtt import Mention

from .. import puppet as pu, user as u
from ..db import Message as DBMessage


class SendParams(NamedTuple):
    text: str
    mentions: List[Mention]
    reply_to: str


class FacebookFormatString(EntityString[SimpleEntity, EntityType], MarkdownString):
    def format(self, entity_type: EntityType, **kwargs) -> 'FacebookFormatString':
        prefix = suffix = ""
        if entity_type == EntityType.USER_MENTION:
            self.entities.append(SimpleEntity(type=entity_type, offset=0, length=len(self.text),
                                              extra_info={"user_id": kwargs["user_id"]}))
            return self
        elif entity_type == EntityType.BOLD:
            prefix = suffix = "*"
        elif entity_type == EntityType.ITALIC:
            prefix = suffix = "_"
        elif entity_type == EntityType.STRIKETHROUGH:
            prefix = suffix = "~"
        elif entity_type == EntityType.URL:
            if kwargs['url'] != self.text:
                suffix = f" ({kwargs['url']})"
        elif entity_type == EntityType.PREFORMATTED:
            prefix = f"```{kwargs['language']}\n"
            suffix = "\n```"
        elif entity_type == EntityType.INLINE_CODE:
            prefix = suffix = "`"
        elif entity_type == EntityType.BLOCKQUOTE:
            children = self.trim().split("\n")
            children = [child.prepend("> ") for child in children]
            return self.join(children, "\n")
        elif entity_type == EntityType.HEADER:
            prefix = "#" * kwargs["size"] + " "
        else:
            return self

        self._offset_entities(len(prefix))
        self.text = f"{prefix}{self.text}{suffix}"
        return self


class MatrixParser(BaseMatrixParser[FacebookFormatString]):
    fs = FacebookFormatString

    @classmethod
    def parse(cls, data: str) -> FacebookFormatString:
        return cast(FacebookFormatString, super().parse(data))


async def matrix_to_facebook(content: TextMessageEventContent, room_id: RoomID) -> SendParams:
    mentions = []
    reply_to = None
    if content.relates_to.rel_type == RelationType.REPLY:
        message = await DBMessage.get_by_mxid(content.relates_to.event_id, room_id)
        if message:
            content.trim_reply_fallback()
            reply_to = message.fbid
    if content.format == Format.HTML and content.formatted_body:
        parsed = MatrixParser.parse(content.formatted_body)
        text = parsed.text
        mentions = []
        for mention in parsed.entities:
            mxid = mention.extra_info["user_id"]
            user = await u.User.get_by_mxid(mxid, create=False)
            if user and user.fbid:
                fbid = user.fbid
            else:
                puppet = await pu.Puppet.get_by_mxid(mxid, create=False)
                if puppet:
                    fbid = puppet.fbid
                else:
                    continue
            mentions.append(Mention(user_id=str(fbid), offset=mention.offset,
                                    length=mention.length))
    else:
        text = content.body
    return SendParams(text=text, mentions=mentions, reply_to=reply_to)



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List, Any, Optional

from mautrix.types import SerializableAttrs
from attr import dataclass


@dataclass
class LoginResponse(SerializableAttrs['LoginResponse']):
    session_key: str
    uid: int
    secret: str
    access_token: str
    machine_id: str
    session_cookies: List[Dict[str, Any]]
    analytics_claim: str
    user_storage_key: str
    identifier: Optional[str] = None


@dataclass
class MobileConfigField(SerializableAttrs['MobileConfigField']):
    k: int
    bln: Optional[int] = None
    i64: Optional[int] = None
    str: Optional[str] = None
    pname: Optional[str] = None


@dataclass
class MobileConfigItem(SerializableAttrs['MobileConfigItem']):
    fields: List[MobileConfigField]
    hash: str


@dataclass
class MobileConfig(SerializableAttrs['MobileConfig']):
    configs: Dict[str, MobileConfigItem]
    query_hash: Optional[str]
    one_query_hash: Optional[str]
    ts: int
    ep_hash: str

    def find(self, number: int, field_k: int) -> Optional[MobileConfigField]:
        try:
            config = self.configs[str(number)]
        except IndexError:
            pass
        else:
            for field in config.fields:
                if field.k == field_k:
                    return field
        return None


@dataclass
class PasswordKeyResponse(SerializableAttrs['PasswordKeyResponse']):
    public_key: str
    key_id: int
    seconds_to_live: int



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Optional, Dict, TypeVar, Type, List, Union
from urllib.parse import quote
import hashlib
import logging
import base64
import random
import json
import time

from mautrix.types import JSON
from aiohttp import ClientSession, ClientResponse
from aiohttp.client import _RequestContextManager
from mautrix.util.logging import TraceLogger
from yarl import URL

from ..state import AndroidState
from ..types import GraphQLQuery, GraphQLMutation
from .errors import ResponseError, error_class_map, error_code_map

T = TypeVar('T')


class BaseAndroidAPI:
    a_url = URL("https://api.facebook.com")
    b_url = URL("https://b-api.facebook.com")
    graph_url = URL("https://graph.facebook.com")
    b_graph_url = URL("https://b-graph.facebook.com")
    rupload_url = URL("https://rupload.facebook.com")
    http: ClientSession
    log: TraceLogger

    # Seems to be a per-minute request identifier
    _cid: str
    _cid_ts: int
    freeze_cid: bool
    # Seems to be a per-session request identifier
    nid: str
    # Seems to be a per-request incrementing integer
    _tid: int

    def __init__(self, state: AndroidState, log: Optional[TraceLogger] = None) -> None:
        self.http = ClientSession()
        self.state = state
        self.log = log or logging.getLogger("mauigpapi.http")
        self._cid = None
        self._cid_ts = 0
        self.freeze_cid = False
        self.nid = base64.b64encode(bytes([random.getrandbits(8) for _ in range(9)])
                                    ).decode("utf-8")
        self._tid = 0
        self._file_url_cache = {}

    @property
    def tid(self) -> int:
        self._tid += 1
        return self._tid

    @property
    def cid(self) -> str:
        new_ts = int(time.time() / 60)
        if not self._cid or (self._cid_ts != new_ts and not self.freeze_cid):
            self._cid_ts = new_ts
            rand = random.Random(f"{self.state.device.uuid}{new_ts}")
            self._cid = bytes([rand.getrandbits(8) for _ in range(16)]).hex()
        return self._cid

    @property
    def session_id(self) -> str:
        return f"nid={self.nid};pid=Main;tid={self.tid};nc=0;fc=0;bc=0,cid={self.cid}"

    def format(self, req: Dict[str, str], sign: bool = True, **extra: str) -> str:
        req = dict(sorted(req.items()))
        if sign:
            sig_data = "".join(f"{key}={value}" for key, value in req.items())
            sig_data_bytes = (sig_data + self.state.application.client_secret).encode("utf-8")
            req["sig"] = hashlib.md5(sig_data_bytes).hexdigest()
        if extra:
            req.update(extra)
        return "&".join(f"{quote(key)}={quote(value)}" for key, value in sorted(req.items()))

    @property
    def _headers(self) -> Dict[str, str]:
        headers = {
            "x-fb-connection-quality": self.state.device.connection_quality,
            "x-fb-connection-type": self.state.device.connection_type,
            "user-agent": self.state.user_agent,
            "x-tigon-is-retry": "False",
            "x-fb-http-engine": "Liger",
            "x-fb-client-ip": "True",
            "x-fb-connection-token": self.cid,
            "x-fb-session-id": self.session_id,
            "x-fb-device-group": self.state.device.device_group,
            "x-fb-sim-hni": str(self.state.carrier.hni),
            "x-fb-net-hni": str(self.state.carrier.hni),
            # "x-fb-background-state": "1",
            "authorization": f"OAuth {self.state.session.access_token or 'null'}",
        }
        return {k: v for k, v in headers.items() if v is not None}

    @property
    def _params(self) -> Dict[str, str]:
        return {
            "locale": self.state.device.language,
            "client_country_code": self.state.device.country_code,
        }

    def get(self, url: Union[str, URL], headers: Optional[Dict[str, str]] = None,
            include_auth: bool = True, **kwargs) -> _RequestContextManager:
        headers = {
            **self._headers,
            **(headers or {}),
        }
        url = URL(url)
        if not url.host.endswith(".facebook.com") or not include_auth:
            headers.pop("authorization")
        return self.http.get(url, headers=headers, **kwargs)

    async def graphql(self, req: GraphQLQuery, headers: Optional[Dict[str, str]] = None,
                      response_type: Optional[Type[T]] = JSON, path: Optional[List[str]] = None,
                      b: bool = True) -> T:
        headers = {
            **self._headers,
            **(headers or {}),
            "content-type": "application/x-www-form-urlencoded",
            "x-fb-friendly-name": req.__class__.__name__,
        }
        variables = req.serialize()
        if isinstance(req, GraphQLMutation):
            variables = {"input": variables}
        params = {
            **self._params,
            "variables": json.dumps(variables),
            "method": "post",
            "doc_id": req.doc_id,
            "format": "json",
            "pretty": "false",
            "query_name": req.__class__.__name__,
            "strip_defaults": "false",
            "strip_nulls": "false",
            "fb_api_req_friendly_name": req.__class__.__name__,
            "fb_api_caller_class": req.caller_class,
        }
        resp = await self.http.post(url=(self.b_graph_url if b else self.graph_url) / "graphql",
                                    data=params, headers=headers)
        self.log.trace(f"GraphQL {req} response: {await resp.text()}")
        if response_type is None:
            self._handle_response_headers(resp)
            return None
        json_data = await self._handle_response(resp)
        if path:
            for item in path:
                json_data = json_data[item]
        if response_type is not JSON:
            return response_type.deserialize(json_data)
        return json_data

    async def _handle_response(self, resp: ClientResponse) -> JSON:
        self._handle_response_headers(resp)
        body = await resp.json()
        error = body.get("error", None)
        if not error:
            return body
        error_class = (error_code_map.get(error["code"])
                       or error_class_map.get(error["type"])
                       or ResponseError)
        raise error_class(error)

    async def _raise_response_error(self, resp: ClientResponse) -> None:
        try:
            data = await resp.json()
        except json.JSONDecodeError:
            data = {}
        # TODO if needed

    def _handle_response_headers(self, resp: ClientResponse) -> None:
        # TODO if needed
        pass



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import (Dict, Deque, Optional, Tuple, Union, Set, AsyncGenerator, List, Any, Awaitable,
                    Pattern, TYPE_CHECKING, cast)
from collections import deque
from html import escape
from io import BytesIO
import mimetypes
import asyncio
import re

from yarl import URL
import magic

from mautrix.types import (RoomID, EventType, ContentURI, MessageEventContent, EventID,
                           ImageInfo, MessageType, LocationMessageEventContent, LocationInfo,
                           ThumbnailInfo, FileInfo, AudioInfo, Format, RelatesTo, RelationType,
                           TextMessageEventContent, MediaMessageEventContent, Membership,
                           EncryptedFile, VideoInfo)
from mautrix.appservice import IntentAPI
from mautrix.errors import MForbidden, IntentError, MatrixError
from mautrix.bridge import BasePortal, NotificationDisabler, async_getter_lock
from mautrix.util.simple_lock import SimpleLock
from mautrix.util.network_retry import call_with_net_retry

from maufbapi.types import mqtt, graphql

from .formatter import facebook_to_matrix, matrix_to_facebook
from .config import Config
from .db import (Portal as DBPortal, Message as DBMessage, Reaction as DBReaction,
                 UserPortal as UserPortal, ThreadType)
from . import puppet as p, user as u

if TYPE_CHECKING:
    from .__main__ import MessengerBridge
    from .matrix import MatrixHandler

try:
    from PIL import Image
except ImportError:
    Image = None

try:
    from mautrix.crypto.attachments import decrypt_attachment, encrypt_attachment
except ImportError:
    decrypt_attachment = encrypt_attachment = None

geo_uri_regex: Pattern = re.compile(r"^geo:(-?\d+.\d+),(-?\d+.\d+)$")


class FakeLock:
    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, exc_type, exc, tb) -> None:
        pass


StateBridge = EventType.find("m.bridge", EventType.Class.STATE)
StateHalfShotBridge = EventType.find("uk.half-shot.bridge", EventType.Class.STATE)
MediaInfo = Union[FileInfo, VideoInfo, AudioInfo, ImageInfo]


class Portal(DBPortal, BasePortal):
    invite_own_puppet_to_pm: bool = False
    by_mxid: Dict[RoomID, 'Portal'] = {}
    by_fbid: Dict[Tuple[int, int], 'Portal'] = {}
    matrix: 'MatrixHandler'
    config: Config

    _main_intent: Optional[IntentAPI]
    _create_room_lock: asyncio.Lock
    _dedup: Deque[str]
    _oti_dedup: Dict[int, EventID]
    _send_locks: Dict[int, asyncio.Lock]
    _noop_lock: FakeLock = FakeLock()
    _typing: Set['u.User']
    backfill_lock: SimpleLock
    _backfill_leave: Optional[Set[IntentAPI]]

    def __init__(self, fbid: int, fb_receiver: int, fb_type: ThreadType,
                 mxid: Optional[RoomID] = None, name: Optional[str] = None,
                 photo_id: Optional[str] = None, avatar_url: Optional[ContentURI] = None,
                 encrypted: bool = False) -> None:
        super().__init__(fbid, fb_receiver, fb_type, mxid, name, photo_id, avatar_url, encrypted)
        self.log = self.log.getChild(self.fbid_log)

        self._main_intent = None
        self._create_room_lock = asyncio.Lock()
        self._dedup = deque(maxlen=100)
        self._oti_dedup = {}
        self._send_locks = {}
        self._typing = set()

        self.backfill_lock = SimpleLock("Waiting for backfilling to finish before handling %s",
                                        log=self.log)
        self._backfill_leave = None

    @classmethod
    def init_cls(cls, bridge: 'MessengerBridge') -> None:
        BasePortal.bridge = bridge
        cls.az = bridge.az
        cls.config = bridge.config
        cls.loop = bridge.loop
        cls.matrix = bridge.matrix
        cls.invite_own_puppet_to_pm = cls.config["bridge.invite_own_puppet_to_pm"]
        NotificationDisabler.puppet_cls = p.Puppet
        NotificationDisabler.config_enabled = cls.config["bridge.backfill.disable_notifications"]

    # region DB conversion

    async def delete(self) -> None:
        if self.mxid:
            await DBMessage.delete_all_by_room(self.mxid)
        self.by_fbid.pop(self.fbid_full, None)
        self.by_mxid.pop(self.mxid, None)
        await super().delete()

    # endregion
    # region Properties

    @property
    def fbid_full(self) -> Tuple[int, int]:
        return self.fbid, self.fb_receiver

    @property
    def fbid_log(self) -> str:
        if self.is_direct:
            return f"{self.fbid}<->{self.fb_receiver}"
        return str(self.fbid)

    @property
    def mqtt_key(self) -> mqtt.ThreadKey:
        if self.fb_type == ThreadType.USER:
            return mqtt.ThreadKey(other_user_id=self.fbid)
        elif self.fb_type == ThreadType.GROUP:
            return mqtt.ThreadKey(thread_fbid=self.fbid)
        else:
            raise ValueError("Unsupported thread type")

    @property
    def graphql_key(self) -> graphql.ThreadKey:
        if self.fb_type == ThreadType.USER:
            return graphql.ThreadKey(other_user_id=str(self.fbid))
        elif self.fb_type == ThreadType.GROUP:
            return graphql.ThreadKey(thread_fbid=str(self.fbid))
        else:
            raise ValueError("Unsupported thread type")

    @property
    def is_direct(self) -> bool:
        return self.fb_type == ThreadType.USER

    @property
    def main_intent(self) -> IntentAPI:
        if not self._main_intent:
            raise ValueError("Portal must be postinit()ed before main_intent can be used")
        return self._main_intent

    # endregion
    # region Chat info updating

    async def update_info(self, source: Optional['u.User'] = None,
                          info: Optional[graphql.Thread] = None) -> Optional[graphql.Thread]:
        if not info:
            self.log.debug("Called update_info with no info, fetching thread info...")
            threads = await source.client.fetch_thread_info(self.fbid)
            if not threads:
                return None
            elif threads[0].thread_key.id != self.fbid:
                self.log.warning("fetch_thread_info response contained different ID (%s) "
                                 "than expected (%s)", threads[0].thread_key.id, self.fbid)
                self.log.debug(f"Number of threads in unexpected response: {len(threads)}")
            info = threads[0]
        if info.thread_key != self.graphql_key:
            self.log.warning("Got different ID (%s) than what asked for (%s) when fetching info",
                             info.thread_key.id, self.fbid)
        changed = False
        if not self.is_direct:
            changed = any(await asyncio.gather(self._update_name(info.name),
                                               self._update_photo(source, info.image),
                                               loop=self.loop))
        changed = await self._update_participants(source, info) or changed
        if changed:
            await self.update_bridge_info()
            await self.save()
        return info

    @staticmethod
    def get_photo_id(photo: Optional[Union[graphql.Picture, str]]) -> Optional[str]:
        if not photo:
            return None
        elif isinstance(photo, graphql.Picture):
            photo = photo.uri
        path = URL(photo).path
        return path[path.rfind("/") + 1:]

    @classmethod
    async def _reupload_fb_file(cls, url: str, source: 'u.User', intent: IntentAPI, *,
                                filename: Optional[str] = None, encrypt: bool = False,
                                referer: str = "messenger_thread_photo", find_size: bool = False,
                                ) -> Tuple[ContentURI, MediaInfo, Optional[EncryptedFile]]:
        if not url:
            raise ValueError("URL not provided")
        headers = {
            "referer": f"fbapp://{source.state.application.client_id}/{referer}"
        }
        async with source.client.get(url, headers=headers) as resp:
            length = int(resp.headers["Content-Length"])
            if length > cls.matrix.media_config.upload_size:
                raise ValueError("File not available: too large")
            data = await resp.read()
        mime = magic.from_buffer(data, mime=True)
        info = FileInfo(mimetype=mime, size=len(data))
        if Image and mime.startswith("image/") and find_size:
            with Image.open(BytesIO(data)) as img:
                width, height = img.size
            info = ImageInfo(mimetype=mime, size=len(data), width=width, height=height)
        upload_mime_type = mime
        decryption_info = None
        if encrypt and encrypt_attachment:
            data, decryption_info = encrypt_attachment(data)
            upload_mime_type = "application/octet-stream"
            filename = None
        url = await call_with_net_retry(intent.upload_media, data, mime_type=upload_mime_type,
                                        filename=filename, _action="upload media")
        if decryption_info:
            decryption_info.url = url
        return url, info, decryption_info

    async def _update_name(self, name: str) -> bool:
        if not name:
            self.log.warning("Got empty name in _update_name call")
            return False
        if self.name != name:
            self.log.trace("Updating name %s -> %s", self.name, name)
            self.name = name
            if self.mxid and (self.encrypted or not self.is_direct):
                await self.main_intent.set_room_name(self.mxid, self.name)
            return True
        return False

    async def _update_photo(self, source: 'u.User', photo: graphql.Picture) -> bool:
        if self.is_direct and not self.encrypted:
            return False
        photo_id = self.get_photo_id(photo)
        if self.photo_id != photo_id:
            self.photo_id = photo_id
            if photo:
                self.avatar_url = await p.Puppet.reupload_avatar(
                    source, self.main_intent, photo.uri,
                    self.fbid, use_graph=self.is_direct and (photo.height or 0) < 500)
            else:
                self.avatar_url = ContentURI("")
            if self.mxid:
                await self.main_intent.set_room_avatar(self.mxid, self.avatar_url)
            return True
        return False

    async def _update_photo_from_puppet(self, puppet: 'p.Puppet') -> bool:
        if self.photo_id == puppet.photo_id:
            return False
        self.photo_id = puppet.photo_id
        if puppet.photo_mxc:
            self.avatar_url = puppet.photo_mxc
        elif self.photo_id:
            profile = await self.main_intent.get_profile(puppet.default_mxid)
            self.avatar_url = profile.avatar_url
            puppet.photo_mxc = profile.avatar_url
        else:
            self.avatar_url = ContentURI("")
        if self.mxid:
            await self.main_intent.set_room_avatar(self.mxid, self.avatar_url)
        return True

    async def _update_participants(self, source: 'u.User', info: graphql.Thread) -> bool:
        changed = False
        for participant in info.all_participants.nodes:
            puppet = await p.Puppet.get_by_fbid(int(participant.id))
            await puppet.update_info(source, participant.messaging_actor)
            if self.is_direct and self.fbid == puppet.fbid and self.encrypted:
                changed = await self._update_name(puppet.name) or changed
                changed = await self._update_photo_from_puppet(puppet) or changed
            if self.mxid and (puppet.fbid != self.fb_receiver or puppet.is_real_user):
                await puppet.intent_for(self).ensure_joined(self.mxid, bot=self.main_intent)
        return changed

    # endregion
    # region Matrix room creation

    async def update_matrix_room(self, source: 'u.User', info: Optional[graphql.Thread] = None
                                 ) -> None:
        try:
            await self._update_matrix_room(source, info)
        except Exception:
            self.log.exception("Failed to update portal")

    async def _update_matrix_room(self, source: 'u.User', info: Optional[graphql.Thread] = None
                                  ) -> None:
        await self.main_intent.invite_user(self.mxid, source.mxid, check_cache=False)
        puppet = await p.Puppet.get_by_custom_mxid(source.mxid)
        if puppet and puppet.is_real_user:
            await puppet.intent.ensure_joined(self.mxid)

        await self.update_info(source, info)

        up = await UserPortal.get(source.fbid, self.fbid, self.fb_receiver)
        if not up:
            in_community = await source._community_helper.add_room(source._community_id, self.mxid)
            await UserPortal(user=source.fbid, portal=self.fbid, portal_receiver=self.fb_receiver,
                             in_community=in_community).insert()
        elif not up.in_community:
            up.in_community = await source._community_helper.add_room(source._community_id,
                                                                      self.mxid)
            await up.save()

    async def create_matrix_room(self, source: 'u.User', info: Optional[graphql.Thread] = None
                                 ) -> Optional[RoomID]:
        if self.mxid:
            try:
                await self._update_matrix_room(source, info)
            except Exception:
                self.log.exception("Failed to update portal")
            return self.mxid
        async with self._create_room_lock:
            try:
                return await self._create_matrix_room(source, info)
            except Exception:
                self.log.exception("Failed to create portal")
                return None

    @property
    def bridge_info_state_key(self) -> str:
        return f"net.maunium.facebook://facebook/{self.fbid}"

    @property
    def bridge_info(self) -> Dict[str, Any]:
        return {
            "bridgebot": self.az.bot_mxid,
            "creator": self.main_intent.mxid,
            "protocol": {
                "id": "facebook",
                "displayname": "Facebook Messenger",
                "avatar_url": self.config["appservice.bot_avatar"],
            },
            "channel": {
                "id": self.fbid,
                "displayname": self.name,
                "avatar_url": self.avatar_url,
            }
        }

    async def update_bridge_info(self) -> None:
        if not self.mxid:
            self.log.debug("Not updating bridge info: no Matrix room created")
            return
        try:
            self.log.debug("Updating bridge info...")
            await self.main_intent.send_state_event(self.mxid, StateBridge,
                                                    self.bridge_info, self.bridge_info_state_key)
            # TODO remove this once https://github.com/matrix-org/matrix-doc/pull/2346 is in spec
            await self.main_intent.send_state_event(self.mxid, StateHalfShotBridge,
                                                    self.bridge_info, self.bridge_info_state_key)
        except Exception:
            self.log.warning("Failed to update bridge info", exc_info=True)

    async def _create_matrix_room(self, source: 'u.User', info: Optional[graphql.Thread] = None
                                  ) -> Optional[RoomID]:
        if self.mxid:
            await self._update_matrix_room(source, info)
            return self.mxid

        self.log.debug(f"Creating Matrix room")
        name: Optional[str] = None
        initial_state = [{
            "type": str(StateBridge),
            "state_key": self.bridge_info_state_key,
            "content": self.bridge_info,
        }, {
            # TODO remove this once https://github.com/matrix-org/matrix-doc/pull/2346 is in spec
            "type": str(StateHalfShotBridge),
            "state_key": self.bridge_info_state_key,
            "content": self.bridge_info,
        }]
        invites = [source.mxid]
        if self.config["bridge.encryption.default"] and self.matrix.e2ee:
            self.encrypted = True
            initial_state.append({
                "type": "m.room.encryption",
                "content": {"algorithm": "m.megolm.v1.aes-sha2"},
            })
            if self.is_direct:
                invites.append(self.az.bot_mxid)

        info = await self.update_info(source=source, info=info)
        if not info:
            self.log.debug("update_info() didn't return info, cancelling room creation")
            return None

        if self.encrypted or not self.is_direct:
            name = self.name
            initial_state.append({"type": str(EventType.ROOM_AVATAR),
                                  "content": {"avatar_url": self.avatar_url}})
        if self.config["appservice.community_id"]:
            initial_state.append({
                "type": "m.room.related_groups",
                "content": {"groups": [self.config["appservice.community_id"]]},
            })

        # We lock backfill lock here so any messages that come between the room being created
        # and the initial backfill finishing wouldn't be bridged before the backfill messages.
        with self.backfill_lock:
            self.mxid = await self.main_intent.create_room(name=name, is_direct=self.is_direct,
                                                           initial_state=initial_state,
                                                           invitees=invites)
            if not self.mxid:
                raise Exception("Failed to create room: no mxid returned")

            if self.encrypted and self.matrix.e2ee and self.is_direct:
                try:
                    await self.az.intent.ensure_joined(self.mxid)
                except Exception:
                    self.log.warning(f"Failed to add bridge bot to new private chat {self.mxid}")

            await self.save()
            self.log.debug(f"Matrix room created: {self.mxid}")
            self.by_mxid[self.mxid] = self
            if not self.is_direct:
                await self._update_participants(source, info)
            else:
                puppet = await p.Puppet.get_by_custom_mxid(source.mxid)
                if puppet:
                    try:
                        did_join = await puppet.intent.join_room_by_id(self.mxid)
                        if did_join and self.fb_type == ThreadType.USER:
                            await source.update_direct_chats({self.main_intent.mxid: [self.mxid]})
                    except MatrixError:
                        self.log.debug("Failed to join custom puppet into newly created portal",
                                       exc_info=True)

            in_community = await source._community_helper.add_room(source._community_id, self.mxid)
            await UserPortal(user=source.fbid, portal=self.fbid, portal_receiver=self.fb_receiver,
                             in_community=in_community).upsert()

            try:
                await self.backfill(source, is_initial=True, thread=info)
            except Exception:
                self.log.exception("Failed to backfill new portal")

        return self.mxid

    # endregion
    # region Matrix event handling

    def require_send_lock(self, user_id: int) -> asyncio.Lock:
        try:
            lock = self._send_locks[user_id]
        except KeyError:
            lock = asyncio.Lock()
            self._send_locks[user_id] = lock
        return lock

    def optional_send_lock(self, user_id: int) -> Union[asyncio.Lock, FakeLock]:
        try:
            return self._send_locks[user_id]
        except KeyError:
            pass
        return self._noop_lock

    async def _send_delivery_receipt(self, event_id: EventID) -> None:
        if event_id and self.config["bridge.delivery_receipts"]:
            try:
                await self.az.intent.mark_read(self.mxid, event_id)
            except Exception:
                self.log.exception("Failed to send delivery receipt for %s", event_id)

    async def _send_bridge_error(self, msg: str) -> None:
        await self._send_message(self.main_intent, TextMessageEventContent(
            msgtype=MessageType.NOTICE,
            body=f"\u26a0 Your message may not have been bridged: {msg}"))

    async def handle_matrix_message(self, sender: 'u.User', message: MessageEventContent,
                                    event_id: EventID) -> None:
        # TODO handle errors?
        # try:
        await self._handle_matrix_message(sender, message, event_id)
        # except fbchat.PleaseRefresh:
        #     self.log.debug(f"Got PleaseRefresh error while trying to bridge {event_id}")
        #     await sender.refresh()
        #     try:
        #         await self._handle_matrix_message(sender, message, event_id)
        #     except fbchat.FacebookError as e:
        #         self.log.exception(f"Got FacebookError while trying to bridge {event_id} "
        #                            "after auto-refreshing")
        #         await self._send_bridge_error(getattr(e, "description", e.message))
        # except fbchat.FacebookError as e:
        #     self.log.exception(f"Got FacebookError while trying to bridge {event_id}")
        #     await self._send_bridge_error(getattr(e, "description", e.message))

    async def _handle_matrix_message(self, sender: 'u.User', message: MessageEventContent,
                                     event_id: EventID) -> None:
        if ((message.get(self.az.real_user_content_key, False)
             and await p.Puppet.get_by_custom_mxid(sender.mxid))):
            self.log.debug(f"Ignoring puppet-sent message by confirmed puppet user {sender.mxid}")
            return
        if message.msgtype == MessageType.TEXT or message.msgtype == MessageType.NOTICE:
            await self._handle_matrix_text(event_id, sender, message)
        elif message.msgtype.is_media:
            await self._handle_matrix_media(event_id, sender, message)
        # elif message.msgtype == MessageType.LOCATION:
        #     await self._handle_matrix_location(sender, message)
        else:
            self.log.warning(f"Unsupported msgtype {message.msgtype} in {event_id}")
            return

    async def _handle_matrix_text(self, event_id: EventID, sender: 'u.User',
                                  message: TextMessageEventContent) -> None:
        converted = await matrix_to_facebook(message, self.mxid)
        offline_threading_id = sender.mqtt.generate_offline_threading_id()
        # TODO store OTI's in the database because sometimes facebook doesn't echo the message back
        #      immediately. Also need to update things like incoming replies to search the database
        #      by OTI instead of only message ID. Outgoing replies may be unsolveable.
        self._oti_dedup[offline_threading_id] = event_id
        resp = await sender.mqtt.send_message(self.fbid, self.fb_type != ThreadType.USER,
                                              message=converted.text, mentions=converted.mentions,
                                              reply_to=converted.reply_to,
                                              offline_threading_id=offline_threading_id)
        if not resp.success and resp.error_message:
            self.log.debug(f"Error handling Matrix message {event_id}: {resp.error_message}")
            await self._send_bridge_error(resp.error_message)
        else:
            self.log.debug(f"Handled Matrix message {event_id} -> OTI: {offline_threading_id}")
            await self._send_delivery_receipt(event_id)

    async def _handle_matrix_media(self, event_id: EventID, sender: 'u.User',
                                   message: MediaMessageEventContent) -> None:
        if message.file and decrypt_attachment:
            data = await self.main_intent.download_media(message.file.url)
            data = decrypt_attachment(data, message.file.key.key,
                                      message.file.hashes.get("sha256"), message.file.iv)
        elif message.url:
            data = await self.main_intent.download_media(message.url)
        else:
            return None
        mime = message.info.mimetype or magic.from_buffer(data, mime=True)
        oti = sender.mqtt.generate_offline_threading_id()
        self._oti_dedup[oti] = event_id
        reply_to = None
        if message.relates_to.rel_type == RelationType.REPLY:
            reply_to_msg = await DBMessage.get_by_mxid(message.relates_to.event_id, self.mxid)
            if reply_to_msg:
                reply_to = reply_to_msg.fbid
        # await sender.mqtt.opened_thread(self.fbid)
        resp = await sender.client.send_media(data, message.body, mime, offline_threading_id=oti,
                                              reply_to=reply_to, chat_id=self.fbid,
                                              is_group=self.fb_type != ThreadType.USER)
        if not resp.media_id and resp.debug_info:
            self.log.debug(f"Error uploading media for Matrix message {event_id}: "
                           f"{resp.debug_info.message}")
            await self._send_bridge_error(f"Media upload error: {resp.debug_info.message}")
            return
        # TODO put message ID from response to database if oti is still on _oti_dedup
        self.log.debug(f"Handled Matrix message {event_id} -> {resp.message_id} / {oti}")
        await self._send_delivery_receipt(event_id)
        # send_resp = await sender.mqtt.send_message(self.fbid, self.fb_type != ThreadType.USER,
        #                                            media_ids=[resp.media_id], reply_to=reply_to,
        #                                            offline_threading_id=oti)
        # if not send_resp.success and send_resp.error_message:
        #     await self._send_bridge_error(send_resp.error_message)

    async def _handle_matrix_location(self, sender: 'u.User',
                                      message: LocationMessageEventContent) -> str:
        pass
        # TODO
        # match = geo_uri_regex.fullmatch(message.geo_uri)
        # return await self.thread_for(sender).send_pinned_location(float(match.group(1)),
        #                                                           float(match.group(2)))

    async def handle_matrix_redaction(self, sender: 'u.User', event_id: EventID,
                                      redaction_event_id: EventID) -> None:
        if not self.mxid:
            return

        message = await DBMessage.get_by_mxid(event_id, self.mxid)
        if message:
            try:
                await message.delete()
                await sender.client.unsend(message.fbid)
                await self._send_delivery_receipt(redaction_event_id)
            except Exception:
                self.log.exception("Unsend failed")
            return

        reaction = await DBReaction.get_by_mxid(event_id, self.mxid)
        if reaction:
            try:
                await message.delete()
                await sender.client.react(message.fbid, None)
                await self._send_delivery_receipt(redaction_event_id)
            except Exception:
                self.log.exception("Removing reaction failed")

    async def handle_matrix_reaction(self, sender: 'u.User', event_id: EventID,
                                     reacting_to: EventID, reaction: str) -> None:
        # Facebook doesn't use variation selectors, Matrix does
        reaction = reaction.rstrip("\ufe0f")

        async with self.require_send_lock(sender.fbid):
            message = await DBMessage.get_by_mxid(reacting_to, self.mxid)
            if not message:
                self.log.debug(f"Ignoring reaction to unknown event {reacting_to}")
                return

            existing = await DBReaction.get_by_fbid(message.fbid, self.fb_receiver, sender.fbid)
            if existing and existing.reaction == reaction:
                return

            await sender.client.react(message.fbid, reaction)
            await self._upsert_reaction(existing, self.main_intent, event_id, message, sender,
                                        reaction)
        await self._send_delivery_receipt(event_id)

    async def handle_matrix_leave(self, user: 'u.User') -> None:
        if self.is_direct:
            self.log.info(f"{user.mxid} left private chat portal with {self.fbid}")
            if user.fbid == self.fb_receiver:
                self.log.info(f"{user.mxid} was the recipient of this portal. "
                              "Cleaning up and deleting...")
                await self.cleanup_and_delete()
        else:
            self.log.debug(f"{user.mxid} left portal to {self.fbid}")

    async def handle_matrix_typing(self, users: Set['u.User']) -> None:
        # FIXME
        pass
        # stopped_typing = [self.thread_for(user).stop_typing() for user in self._typing - users]
        # started_typing = [self.thread_for(user).start_typing() for user in users - self._typing]
        # self._typing = users
        # await asyncio.gather(*stopped_typing, *started_typing, loop=self.loop)

    async def enable_dm_encryption(self) -> bool:
        ok = await super().enable_dm_encryption()
        if ok:
            try:
                puppet = await p.Puppet.get_by_fbid(self.fbid)
                await self.main_intent.set_room_name(self.mxid, puppet.name)
            except Exception:
                self.log.warning(f"Failed to set room name", exc_info=True)
        return ok

    # endregion
    # region Facebook event handling

    async def _bridge_own_message_pm(self, source: 'u.User', sender: 'p.Puppet', mid: str,
                                     invite: bool = True) -> bool:
        if self.is_direct and sender.fbid == source.fbid and not sender.is_real_user:
            if self.invite_own_puppet_to_pm and invite:
                await self.main_intent.invite_user(self.mxid, sender.mxid)
            elif await self.az.state_store.get_membership(self.mxid,
                                                          sender.mxid) != Membership.JOIN:
                self.log.warning(f"Ignoring own {mid} in private chat because own puppet is not in"
                                 " room.")
                return False
        return True

    async def _add_facebook_reply(self, content: TextMessageEventContent, reply: str) -> None:
        if reply:
            message = await DBMessage.get_by_fbid(reply, self.fb_receiver)
            if message:
                evt = await self.main_intent.get_event(message.mx_room, message.mxid)
                if evt:
                    if isinstance(evt.content, TextMessageEventContent):
                        evt.content.trim_reply_fallback()
                    content.set_reply(evt)

    async def _get_facebook_reply(self, reply: str) -> Optional[RelatesTo]:
        if reply:
            message = await DBMessage.get_by_fbid(reply, self.fb_receiver)
            if message:
                return RelatesTo(rel_type=RelationType.REPLY, event_id=message.mxid)
        return None

    async def _send_message(self, intent: IntentAPI, content: MessageEventContent,
                            event_type: EventType = EventType.ROOM_MESSAGE, **kwargs) -> EventID:
        if self.encrypted and self.matrix.e2ee:
            if intent.api.is_real_user:
                content[intent.api.real_user_content_key] = True
            event_type, content = await self.matrix.e2ee.encrypt(self.mxid, event_type, content)
        return await intent.send_message_event(self.mxid, event_type, content, **kwargs)

    async def handle_facebook_message(self, source: 'u.User', sender: 'p.Puppet',
                                      message: Union[graphql.Message, mqtt.Message],
                                      reply_to: Optional[mqtt.Message] = None) -> None:
        try:
            await self._handle_facebook_message(source, sender, message, reply_to)
        except Exception:
            self.log.exception("Error handling Facebook message %s",
                               message.message_id if isinstance(message, graphql.Message)
                               else message.metadata.id)

    async def _handle_facebook_message(self, source: 'u.User', sender: 'p.Puppet',
                                       message: Union[graphql.Message, mqtt.Message],
                                       reply_to: Optional[mqtt.Message] = None) -> None:
        if isinstance(message, graphql.Message):
            self.log.trace("Facebook GraphQL event content: %s", message)
            msg_id = message.message_id
            oti = int(message.offline_threading_id)
            timestamp = message.timestamp
        elif isinstance(message, mqtt.Message):
            self.log.trace("Facebook MQTT event content: %s", message)
            msg_id = message.metadata.id
            oti = message.metadata.offline_threading_id
            timestamp = message.metadata.timestamp
        else:
            raise ValueError(f"Invalid message class {type(message).__name__}")
        if oti in self._oti_dedup:
            event_id = self._oti_dedup.pop(oti)
            self.log.debug("Got message ID %s for offline threading ID %s / %s", msg_id, oti,
                           event_id)
            self._dedup.appendleft(msg_id)
            await DBMessage(mxid=event_id, mx_room=self.mxid, fbid=msg_id, fb_chat=self.fbid,
                            fb_receiver=self.fb_receiver, index=0, timestamp=timestamp).insert()
            return
        elif msg_id in self._dedup:
            self.log.trace("Not handling message %s, found ID in dedup queue", msg_id)
            return
        elif self.backfill_lock.locked and await DBMessage.get_by_fbid(msg_id, self.fb_receiver):
            self.log.trace("Not handling message %s, found duplicate in database", msg_id)
            return
        self.log.debug("Handling Facebook event %s (/%s)", msg_id, oti)
        self._dedup.appendleft(msg_id)
        if not self.mxid:
            mxid = await self.create_matrix_room(source)
            if not mxid:
                # Failed to create
                return
        if not await self._bridge_own_message_pm(source, sender, f"message {msg_id}"):
            return
        intent = sender.intent_for(self)
        if ((self._backfill_leave is not None and self.fbid != sender.fbid
             and intent != sender.intent and intent not in self._backfill_leave)):
            self.log.debug("Adding %s's default puppet to room for backfilling", sender.mxid)
            await self.main_intent.invite_user(self.mxid, intent.mxid)
            await intent.ensure_joined(self.mxid)
            self._backfill_leave.add(intent)
        if isinstance(message, graphql.Message):
            event_ids = await self._handle_graphql_message(source, intent, message)
        else:
            event_ids = await self._handle_mqtt_message(source, intent, message, reply_to)
        if not event_ids:
            self.log.warning(f"Unhandled Messenger message {msg_id}")
            return
        event_ids = [event_id for event_id in event_ids if event_id]
        self.log.debug(f"Handled Messenger message {msg_id} -> {event_ids}")
        await DBMessage.bulk_create(fbid=msg_id, fb_chat=self.fbid, fb_receiver=self.fb_receiver,
                                    mx_room=self.mxid, timestamp=timestamp, event_ids=event_ids)
        await self._send_delivery_receipt(event_ids[-1])

    async def _handle_mqtt_message(self, source: 'u.User', intent: IntentAPI,
                                   message: mqtt.Message, reply_to: Optional[mqtt.Message]
                                   ) -> List[EventID]:
        reply_to_id = reply_to.metadata.id if reply_to else None
        event_ids = []
        if message.sticker:
            event_ids.append(await self._handle_facebook_sticker(
                source, intent, message.sticker, reply_to_id, message.metadata.timestamp
            ))
        if len(message.attachments) > 0:
            attach_ids = await asyncio.gather(
                *[self._handle_facebook_attachment(message.metadata.id, source, intent, attachment,
                                                   reply_to_id, message.metadata.timestamp)
                  for attachment in message.attachments]
            )
            event_ids += [attach_id for attach_id in attach_ids if attach_id]
        if message.text:
            event_ids.append(await self._handle_facebook_text(intent, message, reply_to_id,
                                                              message.metadata.timestamp))
        return event_ids

    async def _convert_extensible_media(self, source: 'u.User', intent: IntentAPI,
                                        sa: graphql.StoryAttachment
                                        ) -> Optional[MessageEventContent]:
        if sa.target.typename == graphql.AttachmentType.EXTERNAL_URL:
            target_url = escape(str(sa.clean_url))
            html = f'<a href="{target_url}">{target_url}</a>'
            return TextMessageEventContent(msgtype=MessageType.TEXT, format=Format.HTML,
                                           body=str(sa.clean_url), formatted_body=html)
        elif sa.media:
            msgtype = {
                "Image": MessageType.IMAGE,
                "Video": MessageType.VIDEO,
            }.get(sa.media.typename_str)
            if sa.media.playable_url and msgtype == MessageType.VIDEO:
                info = VideoInfo()
                url = sa.media.playable_url
            elif sa.media.image_natural and msgtype == MessageType.IMAGE:
                url = sa.media.image_natural.uri
                info = ImageInfo(width=sa.media.image_natural.width,
                                 height=sa.media.image_natural.height)
            else:
                self.log.debug("Unsupported story media attachment: %s",
                               sa.serialize())
                body = "Unsupported shared media attachment"
                html = body
                if sa.title:
                    body = f"{body}: **{sa.title}**"
                    html = f"{html}: <strong>{escape(sa.title)}</strong>"
                if sa.description:
                    body = f"{body}\n\n>{sa.description.text}"
                    html = (f"<p>{html}</p>"
                            f"<blockquote>{escape(sa.description.text)}</blockquote>")
                return TextMessageEventContent(msgtype=MessageType.TEXT, format=Format.HTML,
                                               external_url=sa.url, body=body,
                                               formatted_body=html)
            try:
                mxc, additional_info, decryption_info = await self._reupload_fb_file(
                    url, source, intent, encrypt=self.encrypted, find_size=False)
            except ValueError as e:
                self.log.debug("Failed to reupload story attachment media", exc_info=True)
                return TextMessageEventContent(msgtype=MessageType.NOTICE, body=f"{e}\n{sa.url}",
                                               external_url=sa.url)
            info.size = additional_info.size
            info.mimetype = additional_info.mimetype
            filename = f"{sa.media.typename_str}{mimetypes.guess_extension(info.mimetype)}"
            return MediaMessageEventContent(url=mxc, file=decryption_info, msgtype=msgtype,
                                            body=filename, info=info, external_url=sa.url)
        else:
            self.log.debug("Unhandled story attachment: %s",
                           sa.serialize())
            return None


    async def _convert_mqtt_attachment(self, msg_id: str, source: 'u.User', intent: IntentAPI,
                                       attachment: mqtt.Attachment) -> MessageEventContent:
        filename = attachment.file_name
        if attachment.mime_type and "." not in filename:
            filename += mimetypes.guess_extension(attachment.mime_type)
        referer = "unknown"
        if attachment.extensible_media:
            sa = attachment.parse_extensible().story_attachment
            self.log.trace("Story attachment %s content: %s", attachment.media_id_str, sa)
            return await self._convert_extensible_media(source, intent, sa)
        elif attachment.video_info:
            msgtype = MessageType.VIDEO
            url = attachment.video_info.download_url
            info = VideoInfo(duration=attachment.video_info.duration_ms,
                             width=attachment.video_info.original_width,
                             height=attachment.video_info.original_height)
        elif attachment.audio_info:
            msgtype = MessageType.AUDIO
            url = attachment.audio_info.url
            info = AudioInfo(duration=attachment.audio_info.duration_ms)
        elif attachment.image_info:
            referer = "messenger_thread_photo"
            msgtype = MessageType.IMAGE
            info = ImageInfo(width=attachment.image_info.original_width,
                             height=attachment.image_info.original_height)
            if attachment.image_info.alt_previews:
                url = list(attachment.image_info.alt_previews.values())[0]
                # Override the mime type or detect from file
                attachment.mime_type = {
                    "webp": "image/webp",
                    "gif": "image/gif",
                    "png": "image/png",
                }.get(attachment.image_info.alt_preview_type, None)
            else:
                url = list(attachment.image_info.previews.values())[0]
            # TODO find out if we need to use get_image_url in some cases even with MQTT
            # url = await source.client.get_image_url(msg_id, attachment.media_id)
        elif attachment.media_id:
            # TODO what if it's not a file?
            msgtype = MessageType.FILE
            url = await source.client.get_file_url(self.fbid, msg_id, attachment.media_id)
            info = FileInfo()
        else:
            msg = f"Unsupported attachment"
            self.log.warning(msg)
            return TextMessageEventContent(msgtype=MessageType.NOTICE, body=msg)
        mxc, additional_info, decryption_info = await self._reupload_fb_file(
            url, source, intent, filename=filename, encrypt=self.encrypted,
            find_size=False, referer=referer)
        info.size = additional_info.size
        info.mimetype = attachment.mime_type or additional_info.mimetype
        return MediaMessageEventContent(url=mxc, file=decryption_info, msgtype=msgtype,
                                        body=filename, info=info)

    async def _handle_graphql_message(self, source: 'u.User', intent: IntentAPI,
                                      message: graphql.Message) -> List[EventID]:
        reply_to_id = (message.replied_to_message.message.message_id
                       if message.replied_to_message else None)
        event_ids = []
        if message.sticker:
            event_ids.append(await self._handle_facebook_sticker(
                source, intent, int(message.sticker.id), reply_to_id, message.timestamp
            ))
        if len(message.blob_attachments) > 0:
            attach_ids = await asyncio.gather(
                *[self._handle_facebook_attachment(message.message_id, source, intent, attachment,
                                                   reply_to_id, message.timestamp)
                  for attachment in message.blob_attachments]
            )
            event_ids += [attach_id for attach_id in attach_ids if attach_id]
        if message.extensible_attachment:
            sa = message.extensible_attachment.story_attachment
            content = await self._convert_extensible_media(source, intent, sa)
            if content:
                event_ids.append(await self._send_message(intent, content,
                                                          timestamp=message.timestamp))
        if message.message and message.message.text:
            event_ids.append(await self._handle_facebook_text(intent, message.message, reply_to_id,
                                                              message.timestamp))
        return event_ids

    async def _handle_facebook_text(self, intent: IntentAPI,
                                    message: Union[graphql.MessageText, mqtt.Message],
                                    reply_to_id: str, timestamp: int) -> EventID:
        content = await facebook_to_matrix(message)
        await self._add_facebook_reply(content, reply_to_id)
        return await self._send_message(intent, content, timestamp=timestamp)

    async def _handle_facebook_sticker(self, source: 'u.User', intent: IntentAPI,
                                       sticker_id: int, reply_to: str, timestamp: int) -> EventID:
        resp = await source.client.fetch_stickers([sticker_id], sticker_labels_enabled=True)
        sticker = resp.nodes[0]
        url = (sticker.animated_image or sticker.thread_image).uri
        mxc, info, decryption_info = await self._reupload_fb_file(url, source, intent,
                                                                  encrypt=self.encrypted,
                                                                  find_size=True)
        return await self._send_message(intent, event_type=EventType.STICKER,
                                        content=MediaMessageEventContent(
                                            url=mxc, file=decryption_info, info=info,
                                            msgtype=MessageType.STICKER, body=sticker.label or "",
                                            relates_to=await self._get_facebook_reply(reply_to)),
                                        timestamp=timestamp)

    async def _handle_facebook_attachment(self, msg_id: str, source: 'u.User', intent: IntentAPI,
                                          attachment: Union[graphql.Attachment, mqtt.Attachment],
                                          reply_to: str, timestamp: int) -> Optional[EventID]:
        if isinstance(attachment, graphql.Attachment):
            content = await self._convert_graphql_attachment(msg_id, source, intent, attachment)
        elif isinstance(attachment, mqtt.Attachment):
            content = await self._convert_mqtt_attachment(msg_id, source, intent, attachment)
        else:
            raise ValueError(f"Invalid attachment type {type(attachment).__name__}")
        if not content:
            return None
        content.relates_to = await self._get_facebook_reply(reply_to)
        return await self._send_message(intent, content, timestamp=timestamp)

    async def _convert_graphql_attachment(self, msg_id: str, source: 'u.User', intent: IntentAPI,
                                          attachment: graphql.Attachment) -> MessageEventContent:
        filename = attachment.filename
        if attachment.mimetype and "." not in filename:
            filename += mimetypes.guess_extension(attachment.mimetype)
        referer = "unknown"
        if attachment.typename in (graphql.AttachmentType.IMAGE,
                                   graphql.AttachmentType.ANIMATED_IMAGE):
            msgtype = MessageType.IMAGE
            if attachment.typename == graphql.AttachmentType.IMAGE:
                info = ImageInfo(width=attachment.original_dimensions.x,
                                 height=attachment.original_dimensions.y,
                                 mimetype=attachment.mimetype)
                full_screen = attachment.image_full_screen
            else:
                info = ImageInfo(width=attachment.animated_image_original_dimensions.x,
                                 height=attachment.animated_image_original_dimensions.y,
                                 mimetype=attachment.mimetype)
                full_screen = attachment.animated_image_full_screen
            url = full_screen.uri
            if (info.width, info.height) > full_screen.dimensions:
                url = await source.client.get_image_url(msg_id, attachment.attachment_fbid) or url
            referer = "messenger_thread_photo"
        elif attachment.typename == graphql.AttachmentType.AUDIO:
            msgtype = MessageType.AUDIO
            info = AudioInfo(duration=attachment.playable_duration_in_ms,
                             mimetype=attachment.mimetype)
            url = attachment.playable_url
        elif attachment.typename == graphql.AttachmentType.VIDEO:
            msgtype = MessageType.VIDEO
            info = VideoInfo(duration=attachment.playable_duration_in_ms,
                             mimetype=attachment.mimetype)
            url = attachment.attachment_video_url
        elif attachment.typename == graphql.AttachmentType.FILE:
            msgtype = MessageType.FILE
            url = await source.client.get_file_url(self.fbid, msg_id, attachment.attachment_fbid)
            info = FileInfo(mimetype=attachment.mimetype)
        else:
            # TODO location attachments
            msg = f"Unsupported attachment type {attachment.typename}"
            self.log.warning(msg)
            return TextMessageEventContent(msgtype=MessageType.NOTICE, body=msg)
        mxc, additional_info, decryption_info = await self._reupload_fb_file(
            url, source, intent, filename=filename, encrypt=self.encrypted,
            find_size=False, referer=referer)
        info.size = additional_info.size
        return MediaMessageEventContent(url=mxc, file=decryption_info, msgtype=msgtype,
                                        body=filename, info=info)

    async def _convert_facebook_location(
        self, source: 'u.User', intent: IntentAPI, location: graphql.StoryTarget
    ) -> Union[LocationMessageEventContent, TextMessageEventContent]:
        long, lat = location.coordinates.longitude, location.coordinates.latitude
        if not long or not lat:
            # if location.address or location.url:
            #     self.log.trace("Location message with no coordinates: %s", location)
            #     return TextMessageEventContent(msgtype=MessageType.TEXT,
            #                                    body=f"{location.address}\n{location.url}")
            # else:
            self.log.warning("Unsupported Facebook location message content: %s", location)
            return TextMessageEventContent(msgtype=MessageType.NOTICE,
                                           body="Location message with unsupported content")
        long_char = "E" if long > 0 else "W"
        lat_char = "N" if lat > 0 else "S"
        geo = f"{round(lat, 6)},{round(long, 6)}"

        text = f"{round(abs(lat), 4)}° {lat_char}, {round(abs(long), 4)}° {long_char}"
        url = f"https://maps.google.com/?q={geo}"

        content = LocationMessageEventContent(body=f"Location: {text}\n{url}",
                                              geo_uri=f"geo:{lat},{long}",
                                              msgtype=MessageType.LOCATION)
        # Some clients support formatted body in m.location, so add that as well.
        content["format"] = str(Format.HTML)
        content["formatted_body"] = f"<p>Location: <a href='{url}'>{text}</a></p"
        # TODO find out if locations still have addresses
        # if location.address:
        #     content.body = f"{location.address}\n{content.body}"
        #     content["formatted_body"] = f"<p>{location.address}</p>{content['formatted_body']}"
        return content

    async def handle_facebook_unsend(self, sender: 'p.Puppet', message_id: str, timestamp: int
                                     ) -> None:
        if not self.mxid:
            return
        for message in await DBMessage.get_all_by_fbid(message_id, self.fb_receiver):
            try:
                await sender.intent_for(self).redact(message.mx_room, message.mxid,
                                                     timestamp=timestamp)
            except MForbidden:
                await self.main_intent.redact(message.mx_room, message.mxid,
                                              timestamp=timestamp)
            await message.delete()

    async def handle_facebook_seen(self, source: 'u.User', sender: 'p.Puppet', timestamp: int
                                   ) -> None:
        if not self.mxid:
            return
        msg = await DBMessage.get_closest_before(self.fbid, self.fb_receiver, timestamp)
        if not msg:
            return
        if not await self._bridge_own_message_pm(source, sender, "read receipt",
                                                 invite=False):
            return
        # TODO can we set a timestamp when the read receipt happened?
        await sender.intent_for(self).mark_read(msg.mx_room, msg.mxid)
        self.log.debug(f"Handled Messenger read receipt from {sender.fbid} "
                       f"up to {timestamp}/{msg.mxid}")

    async def handle_facebook_typing(self, source: 'u.User', sender: 'p.Puppet') -> None:
        if not await self._bridge_own_message_pm(source, sender, "typing notification",
                                                 invite=False):
            return
        await sender.intent.set_typing(self.mxid, is_typing=True)

    async def handle_facebook_photo(self, source: 'u.User', sender: 'p.Puppet',
                                    new_photo: mqtt.Attachment, message_id: str, timestamp: int
                                    ) -> None:
        if not self.mxid or self.is_direct or message_id in self._dedup:
            return
        self._dedup.appendleft(message_id)
        photo_url = await source.client.get_image_url(message_id, new_photo.media_id)
        if not photo_url and new_photo.image_info.previews:
            photo_url = list(new_photo.image_info.previews.values())[-1]
        photo_id = self.get_photo_id(photo_url)
        if self.photo_id == photo_id:
            return
        self.photo_id = photo_id
        self.avatar_url, *_ = await self._reupload_fb_file(photo_url, source, sender.intent)
        try:
            event_id = await sender.intent.set_room_avatar(self.mxid, self.avatar_url)
        except IntentError:
            event_id = await self.main_intent.set_room_avatar(self.mxid, self.avatar_url)
        await DBMessage(mxid=event_id, mx_room=self.mxid, index=0, timestamp=timestamp,
                        fbid=message_id, fb_chat=self.fbid, fb_receiver=self.fb_receiver).insert()
        await self.update_bridge_info()

    async def handle_facebook_name(self, source: 'u.User', sender: 'p.Puppet', new_name: str,
                                   message_id: str, timestamp: int) -> None:
        if self.name == new_name or message_id in self._dedup:
            return
        self._dedup.appendleft(message_id)
        self.name = new_name
        if not self.mxid or self.is_direct:
            return
        try:
            event_id = await sender.intent.set_room_name(self.mxid, self.name)
        except IntentError:
            event_id = await self.main_intent.set_room_name(self.mxid, self.name)
        await DBMessage(mxid=event_id, mx_room=self.mxid, index=0, timestamp=timestamp,
                        fbid=message_id, fb_chat=self.fbid, fb_receiver=self.fb_receiver).insert()
        await self.update_bridge_info()

    async def handle_facebook_reaction_add(self, source: 'u.User', sender: 'p.Puppet',
                                           message_id: str, reaction: str) -> None:
        dedup_id = f"react_{message_id}_{sender}_{reaction}"
        async with self.optional_send_lock(sender.fbid):
            if dedup_id in self._dedup:
                return
            self._dedup.appendleft(dedup_id)

        existing = await DBReaction.get_by_fbid(message_id, self.fb_receiver, sender.fbid)
        if existing and existing.reaction == reaction:
            return

        if not await self._bridge_own_message_pm(source, sender, f"reaction to {message_id}"):
            return

        intent = sender.intent_for(self)

        message = await DBMessage.get_by_fbid(message_id, self.fb_receiver)
        if not message:
            self.log.debug(f"Ignoring reaction to unknown message {message_id}")
            return

        matrix_reaction = reaction
        # TODO there are probably other emojis that need variation selectors
        if reaction in ("\u2764", "\U0001f44d", "\U0001f44e"):
            matrix_reaction += "\ufe0f"
        mxid = await intent.react(message.mx_room, message.mxid, matrix_reaction)
        self.log.debug(f"Reacted to {message.mxid}, got {mxid}")

        await self._upsert_reaction(existing, intent, mxid, message, sender, reaction)

    async def _upsert_reaction(self, existing: DBReaction, intent: IntentAPI, mxid: EventID,
                               message: DBMessage, sender: Union['u.User', 'p.Puppet'],
                               reaction: str) -> None:
        if existing:
            self.log.debug(f"_upsert_reaction redacting {existing.mxid} and inserting {mxid}"
                           f" (message: {message.mxid})")
            await intent.redact(existing.mx_room, existing.mxid)
            existing.reaction = reaction
            existing.mxid = mxid
            existing.mx_room = message.mx_room
            await existing.save()
        else:
            self.log.debug(f"_upsert_reaction inserting {mxid} (message: {message.mxid})")
            await DBReaction(mxid=mxid, mx_room=message.mx_room, fb_msgid=message.fbid,
                             fb_receiver=self.fb_receiver, fb_sender=sender.fbid,
                             reaction=reaction).insert()

    async def handle_facebook_reaction_remove(self, source: 'u.User', sender: 'p.Puppet',
                                              message_id: str) -> None:
        if not self.mxid:
            return
        reaction = await DBReaction.get_by_fbid(message_id, self.fb_receiver, sender.fbid)
        if reaction:
            try:
                await sender.intent_for(self).redact(reaction.mx_room, reaction.mxid)
            except MForbidden:
                await self.main_intent.redact(reaction.mx_room, reaction.mxid)
            await reaction.delete()

    async def handle_facebook_join(self, source: 'u.User', sender: 'p.Puppet',
                                   users: List['p.Puppet']) -> None:
        sender_intent = sender.intent_for(self)
        for user in users:
            await sender_intent.invite_user(self.mxid, user.mxid)
            await user.intent_for(self).join_room_by_id(self.mxid)

    async def handle_facebook_leave(self, source: 'u.User', sender: 'p.Puppet', removed: 'p.Puppet'
                                    ) -> None:
        if sender == removed:
            await removed.intent_for(self).leave_room(self.mxid)
        else:
            try:
                await sender.intent_for(self).kick_user(self.mxid, removed.mxid)
            except MForbidden:
                await self.main_intent.kick_user(self.mxid, removed.mxid,
                                                 reason=f"Kicked by {sender.name}")

    # endregion

    async def backfill(self, source: 'u.User', is_initial: bool, thread: graphql.Thread) -> None:
        limit = (self.config["bridge.backfill.initial_limit"] if is_initial
                 else self.config["bridge.backfill.missed_limit"])
        if limit == 0:
            return
        elif limit < 0:
            limit = None
        last_active = None
        if not is_initial and thread and len(thread.last_message.nodes) > 0:
            last_active = thread.last_message.nodes[0].timestamp
        most_recent = await DBMessage.get_most_recent(self.fbid, self.fb_receiver)
        if most_recent and is_initial:
            self.log.debug("Not backfilling %s: already bridged messages found", self.fbid_log)
        elif (not most_recent or not most_recent.timestamp) and not is_initial:
            self.log.debug("Not backfilling %s: no most recent message found", self.fbid_log)
        elif last_active and most_recent.timestamp >= last_active:
            self.log.debug("Not backfilling %s: last activity is equal to most recent bridged "
                           "message (%s >= %s)", self.fbid_log, most_recent.timestamp, last_active)
        else:
            with self.backfill_lock:
                await self._backfill(source, limit, most_recent.timestamp if most_recent else None,
                                     thread=thread)

    async def _backfill(self, source: 'u.User', limit: int, after_timestamp: Optional[int],
                        thread: graphql.Thread) -> None:
        self.log.debug("Backfilling history through %s", source.mxid)
        messages = thread.messages.nodes
        oldest_message = messages[0]
        before_timestamp = oldest_message.timestamp - 1
        self.log.debug("Fetching up to %d messages through %s", limit, source.fbid)
        while len(messages) < limit:
            resp = await source.client.fetch_messages(self.fbid, before_timestamp)
            if not resp.nodes:
                self.log.debug("Stopping fetching messages at %s after empty response",
                               oldest_message.message_id)
                break
            oldest_message = resp.nodes[0]
            before_timestamp = oldest_message.timestamp - 1
            messages = resp.nodes + messages
            if not resp.page_info.has_previous_page:
                self.log.debug("Stopping fetching messages at %s as response said there are no "
                               "more messages", oldest_message.message_id)
                break
            elif after_timestamp and oldest_message.timestamp <= after_timestamp:
                self.log.debug("Stopping fetching messages at %s as message is older than newest "
                               "bridged message (%s < %s)", oldest_message.message_id,
                               oldest_message.timestamp, after_timestamp)
                break
        if after_timestamp:
            try:
                slice_index = next(index for index, message in enumerate(messages)
                                   if message.timestamp > after_timestamp)
                messages = messages[slice_index:]
            except StopIteration:
                messages = []
        if not messages:
            self.log.debug("Didn't get any messages from server")
            return
        self.log.debug("Got %d messages from server", len(messages))
        self._backfill_leave = set()
        async with NotificationDisabler(self.mxid, source):
            for message in messages:
                puppet = await p.Puppet.get_by_fbid(message.message_sender.id)
                await self.handle_facebook_message(source, puppet, message)
        for intent in self._backfill_leave:
            self.log.trace("Leaving room with %s post-backfill", intent.mxid)
            await intent.leave_room(self.mxid)
        self.log.info("Backfilled %d messages through %s", len(messages), source.mxid)

    # region Database getters

    async def postinit(self) -> None:
        self.by_fbid[self.fbid_full] = self
        if self.mxid:
            self.by_mxid[self.mxid] = self
        self._main_intent = ((await p.Puppet.get_by_fbid(self.fbid)).default_mxid_intent
                             if self.is_direct else self.az.intent)

    @classmethod
    @async_getter_lock
    async def get_by_mxid(cls, mxid: RoomID) -> Optional['Portal']:
        try:
            return cls.by_mxid[mxid]
        except KeyError:
            pass

        portal = cast(cls, await super().get_by_mxid(mxid))
        if portal:
            await portal.postinit()
            return portal

        return None

    @classmethod
    @async_getter_lock
    async def get_by_fbid(cls, fbid: int, *, fb_receiver: int = 0, create: bool = True,
                          fb_type: Optional[ThreadType] = None) -> Optional['Portal']:
        if fb_type:
            fb_receiver = fb_receiver if fb_type == ThreadType.USER else 0
        fbid_full = (fbid, fb_receiver)
        try:
            return cls.by_fbid[fbid_full]
        except KeyError:
            pass

        portal = cast(cls, await super().get_by_fbid(fbid, fb_receiver))
        if portal:
            await portal.postinit()
            return portal

        if fb_type and create:
            portal = cls(fbid=fbid, fb_receiver=fb_receiver, fb_type=fb_type)
            await portal.insert()
            await portal.postinit()
            return portal

        return None

    @classmethod
    async def get_all_by_receiver(cls, fb_receiver: int) -> AsyncGenerator['Portal', None]:
        portals = await super().get_all_by_receiver(fb_receiver)
        portal: Portal
        for portal in portals:
            try:
                yield cls.by_fbid[(portal.fbid, portal.fb_receiver)]
            except KeyError:
                await portal.postinit()
                yield portal

    @classmethod
    async def all(cls) -> AsyncGenerator['Portal', None]:
        portals = await super().all()
        portal: Portal
        for portal in portals:
            try:
                yield cls.by_fbid[(portal.fbid, portal.fb_receiver)]
            except KeyError:
                await portal.postinit()
                yield portal

    @classmethod
    def get_by_thread(cls, key: Union[graphql.ThreadKey, mqtt.ThreadKey],
                      fb_receiver: Optional[int] = None, create: bool = True
                      ) -> Awaitable['Portal']:
        return cls.get_by_fbid(key.id, fb_receiver=fb_receiver, create=create,
                               fb_type=ThreadType.from_thread_key(key))

    # endregion



"""Add messenger domain to db

Revision ID: 3c59aa2a3af9
Revises: 2162f7ae5365
Create Date: 2020-08-31 18:48:20.989464

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '3c59aa2a3af9'
down_revision = '2162f7ae5365'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.add_column(sa.Column('fb_domain', sa.String(length=255), nullable=False,
                                      server_default="messenger.com"))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('user', schema=None) as batch_op:
        batch_op.drop_column('fb_domain')

    # ### end Alembic commands ###



from .login import LoginResponse, MobileConfig, PasswordKeyResponse
from .common import MessageUnsendability, ThreadFolder
from .graphql import (GraphQLQuery, MoreMessagesQuery, ThreadListQuery, ThreadListResponse,
                      MessageList, GraphQLMutation, ThreadNameMutation, StickerPreviewResponse,
                      FetchStickersWithPreviewsQuery, MessageUndoSend, MessageUnsendResponse,
                      ReactionAction, MessageReactionMutation, DownloadImageFragment,
                      ImageFragment, SubsequentMediaQuery, FbIdToCursorQuery, ThreadQuery,
                      SubsequentMediaResponse, FileAttachmentUrlQuery, FileAttachmentURLResponse,
                      SearchEntitiesResponse, SearchEntitiesNamedQuery, ThreadQueryResponse)
from .mqtt import (RealtimeConfig, RealtimeClientInfo, SendMessageRequest, MessageSyncPayload,
                   MarkReadRequest, OpenedThreadRequest, SendMessageResponse, RegionHintPayload)
from .media import UploadResponse, UploadErrorData



# mautrix-facebook - A Matrix-Facebook Messenger puppeting bridge.
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import List, Dict, Optional, Any
import base64
import json

from attr import dataclass
import attr

from mautrix.types import SerializableAttrs, SerializableEnum, ExtensibleEnum
from maufbapi.thrift import TType, RecursiveType, ThriftObject, field, autospec
from ..common import MessageUnsendability as Unsendability
from ..graphql import ExtensibleAttachment


@autospec
@dataclass
class ThreadKey(ThriftObject):
    other_user_id: int = field(TType.I64, default=None)
    thread_fbid: int = field(TType.I64, default=None)

    @property
    def id(self) -> Optional[int]:
        if self.other_user_id:
            return self.other_user_id
        elif self.thread_fbid:
            return self.thread_fbid
        else:
            return None


@autospec
@dataclass(kw_only=True)
class MessageMetadata(ThriftObject):
    thread: ThreadKey
    id: str
    offline_threading_id: int = field(TType.I64, default=None)
    sender: int = field(TType.I64)
    timestamp: int = field(TType.I64)
    # index 6: unknown bool (ex: true)
    action_summary: str = field(index=7, default=None)
    tags: List[str] = field(factory=lambda: [])
    # index 9: unknown int32 (ex: 3)
    # index 10: unknown bool (ex: false)
    # index 11: ???
    message_unsendability: Unsendability = field(TType.BINARY, index=12,
                                                 default=Unsendability.DENY_FOR_NON_SENDER)
    # indices 13-16: ???
    # index 17: struct (or maybe union?)
    #   index 1: int64 group id
    #   index 2: set of int64 recipient user ids in private chats?


@autospec
@dataclass(kw_only=True)
class ImageInfo(ThriftObject):
    original_width: int = field(TType.I32)
    original_height: int = field(TType.I32)
    previews: Dict[int, str] = field(TType.MAP, key_type=TType.I32, default=None,
                                     value_type=RecursiveType(TType.BINARY, python_type=str))
    # index 4: unknown int32
    # indices 5 and 6: ???
    alt_previews: Dict[int, str] = field(TType.MAP, key_type=TType.I32, default=None, index=7,
                                         value_type=RecursiveType(TType.BINARY, python_type=str))
    image_type: str = field(default=None)
    alt_preview_type: str = field(default=None)
    # index 9: ???
    # index 10: unknown bool


@autospec
@dataclass(kw_only=True)
class VideoInfo(ThriftObject):
    original_width: int = field(TType.I32)
    original_height: int = field(TType.I32)
    duration_ms: int = field(TType.I32)
    thumbnail_url: str
    download_url: str
    # index 6: unknown int32 (ex: 1)
    # index 7: unknown int32 (ex: 0)
    # index 8: unknown int32 (ex: 0)


@autospec
@dataclass(kw_only=True)
class AudioInfo(ThriftObject):
    # index 1: mysterious boolean (true)
    # index 2: mysterious binary (empty)
    url: str = field(index=3)
    duration_ms: int = field(TType.I32)


@autospec
@dataclass(kw_only=True)
class Attachment(ThriftObject):
    media_id_str: str
    mime_type: str = field(default=None)
    file_name: str = field(default=None)
    media_id: int = field(TType.I64, default=None)
    file_size: int = field(TType.I64, default=None)
    # index 6: ???
    extensible_media: str = field(default=None, index=7)
    # indices 8 and 9: ???
    image_info: ImageInfo = field(default=None, index=10)
    video_info: VideoInfo = field(default=None)
    audio_info: AudioInfo = field(default=None)
    # can contain a dash_manifest key with some XML as the value
    # or fbtype key with a number as value
    extra_metadata: Dict[str, str] = field(factory=lambda: {})

    # index 1007?!: unknown bool

    def parse_extensible(self) -> ExtensibleAttachment:
        if not self.extensible_media:
            raise ValueError("This attachment does not contain an extensible attachment")
        data = json.loads(self.extensible_media)
        raw_media_key = f"extensible_message_attachment:{self.media_id_str}"
        expected_key = base64.b64encode(raw_media_key.encode("utf-8")).decode("utf-8").rstrip("=")
        try:
            media_data = data[expected_key]
        except KeyError:
            media_data = list(data.values())[0]
        return ExtensibleAttachment.deserialize(media_data)


@autospec
@dataclass(kw_only=True)
class Reaction(ThriftObject):
    thread: ThreadKey
    message_id: str
    # index 3: unknown int32 (zero)
    reaction_sender_id: int = field(TType.I64, index=4)
    reaction: str = field(default=None)
    message_sender_id: int = field(TType.I64)
    # index 7: unknown number as string, similar to MessageMetadata's index 3


class MentionType(SerializableEnum):
    PERSON = "p"


@dataclass
class Mention(SerializableAttrs['Mention']):
    offset: int = attr.ib(metadata={"json": "o"})
    length: int = attr.ib(metadata={"json": "l"})
    user_id: str = attr.ib(metadata={"json": "i"})
    type: MentionType = attr.ib(metadata={"json": "t"}, default=MentionType.PERSON)


@autospec
@dataclass(kw_only=True)
class Message(ThriftObject):
    metadata: MessageMetadata
    text: str = field(default=None)
    # index 3: ???
    sticker: int = field(TType.I64, index=4, default=None)
    attachments: List[Attachment] = field(factory=lambda: [])
    # index 6: some sort of struct:
    #    1: List[BinaryThreadKey]?
    #    2: ???
    #    3: timestamp?
    #    4: timestamp?
    extra_metadata: Dict[str, bytes] = field(index=7, factory=lambda: {})

    # index 1000?!: int64 (ex: 81)
    # index 1017: int64 (ex: 924)
    # index 1003: struct
    #   index 1: struct
    #     index 1: binary, replying to message id
    # index 1012: map<binary, binary>
    #   key apiArgs: binary containing thrift
    #     index 2: binary url, https://www.facebook.com/intern/agent/realtime_delivery/
    #     index 4: int64 (ex: 0)
    #     index 7: binary, empty?
    #     index 5: binary, some sort of uuid
    #     index 8: list<map>
    #       item 1: map<binary, binary>
    #         {"layer": "www", "push_phase": "C3", "www_rev": "1003179603",
    #          "buenopath": "XRealtimeDeliveryThriftServerController:sendRealtimeDeliveryRequest:/ls_req:TASK_LABEL=SEND_MESSAGE_V{N}"}
    #     index 9: binary (ex: www)
    #     index 10: boolean (ex: false)
    # index 1015: list<binary>, some sort of tags

    @property
    def mentions(self) -> List[Mention]:
        return [Mention.deserialize(item) for item
                in json.loads(self.extra_metadata.get("prng", "[]"))]


@autospec
@dataclass(kw_only=True)
class ExtendedMessage(ThriftObject):
    reply_to_message: Message
    message: Message


@autospec
@dataclass(kw_only=True)
class UnsendMessage(ThriftObject):
    thread: ThreadKey
    message_id: str
    timestamp: int = field(TType.I64)
    user_id: int = field(TType.I64)
    # index 5: unknown int64 (ex: 0)


@autospec
@dataclass
class ExtendedAddMemberParticipant(ThriftObject):
    addee_user_id: int = field(TType.I64)
    adder_user_id: int = field(TType.I64)
    # index 3: unknown int32 (ex: 0)
    timestamp: int = field(TType.I64, index=4)


@autospec
@dataclass
class ExtendedAddMember(ThriftObject):
    thread: ThreadKey
    users: List[ExtendedAddMemberParticipant]


@autospec
@dataclass
class MessageSyncInnerEvent(ThriftObject):
    reaction: Reaction = field(index=10, default=None)
    extended_add_member: ExtendedAddMember = field(index=42, default=None)
    extended_message: ExtendedMessage = field(index=55, default=None)
    unsend_message: UnsendMessage = field(index=67, default=None)


@autospec
@dataclass
class MessageSyncInnerPayload(ThriftObject):
    items: List[MessageSyncInnerEvent]


@autospec
@dataclass(kw_only=True)
class BinaryData(ThriftObject):
    data: bytes

    def parse(self) -> MessageSyncInnerPayload:
        return MessageSyncInnerPayload.from_thrift(self.data)


@autospec
@dataclass
class ReadReceipt(ThriftObject):
    thread: ThreadKey
    user_id: int = field(TType.I64)
    read_at: int = field(TType.I64)
    read_to: int = field(TType.I64)


@autospec
@dataclass
class OwnReadReceipt(ThriftObject):
    threads: List[ThreadKey]
    # index 2: ???
    read_to: int = field(TType.I64, index=3)
    read_at: int = field(TType.I64)


@autospec
@dataclass
class NameChange(ThriftObject):
    metadata: MessageMetadata
    new_name: str


@autospec
@dataclass
class AvatarChange(ThriftObject):
    metadata: MessageMetadata
    new_avatar: Attachment


class ThreadChangeAction(ExtensibleEnum):
    # action_data:
    #   'thread_icon_url': 'https://www.facebook.com/images/emoji.php/v9/t54/1/16/1f408.png'
    #   'thread_icon': '🐈'
    ICON = "change_thread_icon"

    # action_data:
    #   'should_show_icon': '1'
    #   'theme_color': 'FF5E007E'
    #   'accessibility_label': 'Grape'
    THEME = "change_thread_theme"

    # action_data:
    #   'THREAD_CATEGORY': 'GROUP'
    #   'TARGET_ID': '<user id>'
    #   'ADMIN_TYPE': '0'
    #   'ADMIN_EVENT': 'add_admin' or 'remove_admin'
    ADMINS = "change_thread_admins"

    # action_data:
    #   'APPROVAL_MODE': '1' (or '0'?)
    #   'THREAD_CATEGORY': 'GROUP'
    APPROVAL_MODE = "change_thread_approval_mode"


@autospec
@dataclass(kw_only=True)
class ThreadChange(ThriftObject):
    metadata: MessageMetadata
    action: ThreadChangeAction = field(TType.BINARY)
    action_data: Dict[str, str] = field(TType.MAP,
                                        key_type=RecursiveType(TType.BINARY, python_type=str),
                                        value_type=RecursiveType(TType.BINARY, python_type=str))


@autospec
@dataclass(kw_only=True)
class AddMemberParticipant(ThriftObject):
    id: int = field(TType.I64)
    first_name: str
    name: str
    # index 4: unknown boolean


@autospec
@dataclass(kw_only=True)
class AddMember(ThriftObject):
    metadata: MessageMetadata
    users: List[AddMemberParticipant]


@autospec
@dataclass(kw_only=True)
class RemoveMember(ThriftObject):
    metadata: MessageMetadata
    user_id: int = field(TType.I64)


@autospec
@dataclass(kw_only=True)
class MessageSyncEvent(ThriftObject):
    # index 1: unknown struct (no fields known)
    message: Message = field(index=2, default=None)
    own_read_receipt: OwnReadReceipt = field(index=4, default=None)
    add_member: AddMember = field(index=8, default=None)
    remove_member: RemoveMember = field(index=9, default=None)
    name_change: NameChange = field(index=10, default=None)
    avatar_change: AvatarChange = field(index=11, default=None)
    thread_change: ThreadChange = field(index=17, default=None)
    read_receipt: ReadReceipt = field(index=19, default=None)
    # index 25: unknown struct
    #   index 1: ThreadKey
    #   index 2: some user ID
    #   index 6: list of binary (message IDs)
    #   index 7: timestamp
    binary: BinaryData = field(index=42, default=None)

    def get_parts(self) -> List[Any]:
        parts = [self.message, self.own_read_receipt, self.add_member, self.remove_member,
                 self.name_change, self.avatar_change, self.thread_change, self.read_receipt]
        if self.binary:
            for inner_item in self.binary.parse().items:
                parts += [inner_item.reaction, inner_item.extended_message,
                          inner_item.unsend_message, inner_item.extended_add_member]
        return [part for part in parts if part is not None]


class MessageSyncError(ExtensibleEnum):
    QUEUE_OVERFLOW = "ERROR_QUEUE_OVERFLOW"
    QUEUE_UNDERFLOW = "ERROR_QUEUE_UNDERFLOW"


@autospec
@dataclass(kw_only=True)
class MessageSyncPayload(ThriftObject):
    items: List[MessageSyncEvent] = field(factory=lambda: [])
    first_seq_id: int = field(TType.I64, default=None)
    last_seq_id: int = field(TType.I64, default=None)
    viewer: int = field(TType.I64, default=None)
    # indices 5-10: ???
    subscribe_ok: str = field(index=11, default=None)
    error: MessageSyncError = field(TType.BINARY, default=None)


@autospec
@dataclass(kw_only=True)
class SendMessageResponse(ThriftObject):
    offline_threading_id: int = field(TType.I64)
    success: bool
    # index 3: unknown i32 present for errors
    error_message: str = field(default=None, index=4)
    # index 5: unknown boolean present for errors


@autospec
@dataclass
class RegionHint(ThriftObject):
    code: str


@autospec
@dataclass(kw_only=True)
class RegionHintPayload(ThriftObject):
    unknown_int64: int = field(TType.I64)
    region_hint_data: bytes

    @property
    def region_hint(self) -> RegionHint:
        return RegionHint.from_thrift(self.region_hint_data)



"""Store Matrix avatar URL in database

Revision ID: 54aecbd6187c
Revises: 76cb89f6b352
Create Date: 2020-06-15 13:52:37.932842

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '54aecbd6187c'
down_revision = '76cb89f6b352'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('portal', schema=None) as batch_op:
        batch_op.add_column(sa.Column('avatar_url', sa.String(length=255), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('portal', schema=None) as batch_op:
        batch_op.drop_column('avatar_url')

    # ### end Alembic commands ###
